#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Any, Dict, List

import humps


class OnlineConfig:
    """
    Metadata object used to provide online Feature Store configuration information for a feature group.
    """

    def __init__(
        self,
        online_comments: List[str] = None,
        table_space: str = None,
        **kwargs,
    ):
        self._online_comments = online_comments
        self._table_space = table_space

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> "OnlineConfig":
        if json_dict is None:
            return None

        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def to_dict(self):
        return {
            "onlineComments": self._online_comments,
            "tableSpace": self._table_space,
        }

    @property
    def online_comments(self) -> List[str]:
        """List of comments applied to online feature store table."""
        return self._online_comments

    @online_comments.setter
    def online_comments(self, online_comments: List[str]) -> None:
        self._online_comments = online_comments

    @property
    def table_space(self) -> str:
        """Table space of online feature store table for storing data on disk."""
        return self._table_space

    @table_space.setter
    def table_space(self, table_space: str) -> None:
        self._table_space = table_space



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import itertools
import json
import re
import sys
import threading
import time
from datetime import date, datetime, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Literal,
    Optional,
    Tuple,
    Union,
)
from urllib.parse import urljoin, urlparse

from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException, JobException
from hopsworks_common.git_file_status import GitFileStatus


if TYPE_CHECKING:
    import pandas as pd


FEATURE_STORE_NAME_SUFFIX = "_featurestore"


class Encoder(json.JSONEncoder):
    def default(self, o: Any) -> Dict[str, Any]:
        try:
            return o.to_dict()
        except AttributeError:
            return super().default(o)


VALID_EMBEDDING_TYPE = {
    "array<int>",
    "array<bigint>",
    "array<float>",
    "array<double>",
}


def validate_embedding_feature_type(embedding_index, schema):
    if not embedding_index or not schema:
        return
    feature_type_map = dict([(feat.name, feat.type) for feat in schema])
    for embedding in embedding_index.get_embeddings():
        feature_type = feature_type_map.get(embedding.name)
        if feature_type not in VALID_EMBEDDING_TYPE:
            raise FeatureStoreException(
                f"Provide feature `{embedding.name}` has type `{feature_type}`, "
                f"but requires one of the following: {', '.join(VALID_EMBEDDING_TYPE)}"
            )


def autofix_feature_name(name: str) -> str:
    # replace spaces with underscores and enforce lower case
    return name.lower().replace(" ", "_")


def feature_group_name(
    feature_group,  #  FeatureGroup | ExternalFeatureGroup | SpineGroup
) -> str:
    return feature_group.name + "_" + str(feature_group.version)


def append_feature_store_suffix(name: str) -> str:
    name = name.lower()
    if name.endswith(FEATURE_STORE_NAME_SUFFIX):
        return name
    else:
        return name + FEATURE_STORE_NAME_SUFFIX


def strip_feature_store_suffix(name: str) -> str:
    name = name.lower()
    if name.endswith(FEATURE_STORE_NAME_SUFFIX):
        return name[: -1 * len(FEATURE_STORE_NAME_SUFFIX)]
    else:
        return name


def get_dataset_type(path: str) -> Literal["HIVEDB", "DATASET"]:
    if re.match(r"^(?:hdfs://|)/apps/hive/warehouse/*", path):
        return "HIVEDB"
    else:
        return "DATASET"


def check_timestamp_format_from_date_string(input_date: str) -> Tuple[str, str]:
    date_format_patterns = {
        r"^([0-9]{4})([0-9]{2})([0-9]{2})$": "%Y%m%d",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H%M",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H%M%S",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{3})$": "%Y%m%d%H%M%S%f",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})T([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{6})Z$": "ISO",
    }
    normalized_date = (
        input_date.replace("/", "")
        .replace("-", "")
        .replace(" ", "")
        .replace(":", "")
        .replace(".", "")
    )

    date_format = None
    for pattern in date_format_patterns:
        date_format_pattern = re.match(pattern, normalized_date)
        if date_format_pattern:
            date_format = date_format_patterns[pattern]
            break

    if date_format is None:
        raise ValueError(
            "Unable to identify format of the provided date value : " + input_date
        )

    return normalized_date, date_format


def get_timestamp_from_date_string(input_date: str) -> int:
    norm_input_date, date_format = check_timestamp_format_from_date_string(input_date)
    try:
        if date_format != "ISO":
            date_time = datetime.strptime(norm_input_date, date_format)
        else:
            date_time = datetime.fromisoformat(input_date[:-1])
    except ValueError as err:
        raise ValueError(
            "Unable to parse the normalized input date value : "
            + norm_input_date
            + " with format "
            + date_format
        ) from err
    if date_time.tzinfo is None:
        date_time = date_time.replace(tzinfo=timezone.utc)
    return int(float(date_time.timestamp()) * 1000)


def get_hudi_datestr_from_timestamp(timestamp: int) -> str:
    return datetime.fromtimestamp(timestamp / 1000, timezone.utc).strftime(
        "%Y%m%d%H%M%S%f"
    )[:-3]


def get_delta_datestr_from_timestamp(timestamp: int) -> str:
    return datetime.fromtimestamp(timestamp / 1000, timezone.utc).strftime(
        "%Y-%m-%d %H:%M:%S.%f"
    )[:-3]


def convert_event_time_to_timestamp(
    event_time: Optional[
        Union[str, pd._libs.tslibs.timestamps.Timestamp, datetime, date, int]
    ],
) -> Optional[int]:
    if not event_time:
        return None
    if isinstance(event_time, str):
        return get_timestamp_from_date_string(event_time)
    elif hasattr(event_time, "to_pydatetime"):
        # only pandas Timestamp has to_pydatetime method out of the accepted event_time types
        # convert to unix epoch time in milliseconds.
        event_time = event_time.to_pydatetime()
        # convert to unix epoch time in milliseconds.
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, datetime):
        # convert to unix epoch time in milliseconds.
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, date):
        # convert to unix epoch time in milliseconds.
        event_time = datetime(*event_time.timetuple()[:7])
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, int):
        if event_time == 0:
            raise ValueError("Event time should be greater than 0.")
        # jdbc supports timestamp precision up to second only.
        if len(str(event_time)) <= 10:
            event_time = event_time * 1000
        return event_time
    else:
        raise ValueError(
            "Given event time should be in `datetime`, `date`, `str` or `int` type"
        )


def get_hostname_replaced_url(sub_path: str) -> str:
    """
    construct and return an url with public hopsworks hostname and sub path
    :param self:
    :param sub_path: url sub-path after base url
    :return: href url
    """
    href = urljoin(client.get_instance()._base_url, sub_path)
    url_parsed = client.get_instance().replace_public_host(urlparse(href))
    return url_parsed.geturl()


def verify_attribute_key_names(
    feature_group_obj,  #  FeatureGroup | ExternalFeatureGroup | SpineGroup
    external_feature_group: bool = False,
) -> None:
    feature_names = set(feat.name for feat in feature_group_obj.features)
    if feature_group_obj.primary_key:
        diff = set(feature_group_obj.primary_key) - feature_names
        if diff:
            raise FeatureStoreException(
                f"Provided primary key(s) {','.join(diff)} doesn't exist in feature dataframe"
            )

    if feature_group_obj.event_time:
        if feature_group_obj.event_time not in feature_names:
            raise FeatureStoreException(
                f"Provided event_time feature {feature_group_obj.event_time} doesn't exist in feature dataframe"
            )

    if not external_feature_group:
        if feature_group_obj.partition_key:
            diff = set(feature_group_obj.partition_key) - feature_names
            if diff:
                raise FeatureStoreException(
                    f"Provided partition key(s) {','.join(diff)} doesn't exist in feature dataframe"
                )

        if feature_group_obj.hudi_precombine_key:
            if feature_group_obj.hudi_precombine_key not in feature_names:
                raise FeatureStoreException(
                    f"Provided hudi precombine key {feature_group_obj.hudi_precombine_key} "
                    f"doesn't exist in feature dataframe"
                )


def get_job_url(href: str) -> str:
    """Use the endpoint returned by the API to construct the UI url for jobs

    Args:
        href (str): the endpoint returned by the API
    """
    url = urlparse(href)
    url_splits = url.path.split("/")
    project_id = url_splits[4]
    job_name = url_splits[6]
    ui_url = url._replace(
        path="p/{}/jobs/named/{}/executions".format(project_id, job_name)
    )
    ui_url = client.get_instance().replace_public_host(ui_url)
    return ui_url.geturl()


def _loading_animation(message: str, stop_event: threading.Event) -> None:
    for char in itertools.cycle([".", "..", "...", ""]):
        if stop_event.is_set():
            break
        print(f"{message}{char}   ", end="\r")
        time.sleep(0.5)


def run_with_loading_animation(message: str, func: Callable, *args, **kwargs) -> Any:
    stop_event = threading.Event()
    t = threading.Thread(
        target=_loading_animation,
        args=(
            message,
            stop_event,
        ),
    )
    t.daemon = True
    t.start()
    start = time.time()
    end = None

    try:
        result = func(*args, **kwargs)
        end = time.time()
        return result
    finally:
        # Stop the animation and print the "Finished Querying" message
        stop_event.set()
        t.join()
        if not end:
            print(f"\rError: {message}           ", end="\n")
        else:
            print(f"\rFinished: {message} ({(end-start):.2f}s) ", end="\n")


def get_feature_group_url(feature_store_id: int, feature_group_id: int) -> str:
    sub_path = (
        "/p/"
        + str(client.get_instance()._project_id)
        + "/fs/"
        + str(feature_store_id)
        + "/fg/"
        + str(feature_group_id)
    )
    return get_hostname_replaced_url(sub_path)


def is_runtime_notebook():
    if "ipykernel" in sys.modules:
        return True
    else:
        return False


class VersionWarning(Warning):
    pass


class JobWarning(Warning):
    pass


class StorageWarning(Warning):
    pass


class StatisticsWarning(Warning):
    pass


class ValidationWarning(Warning):
    pass


class FeatureGroupWarning(Warning):
    pass


def convert_to_abs(path, current_proj_name):
    abs_project_prefix = "/Projects/{}".format(current_proj_name)
    if not path.startswith(abs_project_prefix):
        return abs_project_prefix + "/" + path
    else:
        return path


def validate_job_conf(config, project_name):
    # User is required to set the appPath programmatically after getting the configuration
    if (
        config["type"] != "dockerJobConfiguration"
        and config["type"] != "flinkJobConfiguration"
        and "appPath" not in config
    ):
        raise JobException("'appPath' not set in job configuration")
    elif "appPath" in config and not config["appPath"].startswith("hdfs://"):
        config["appPath"] = "hdfs://" + convert_to_abs(config["appPath"], project_name)

    # If PYSPARK application set the mainClass, if SPARK validate there is a mainClass set
    if config["type"] == "sparkJobConfiguration":
        if config["appPath"].endswith(".py"):
            config["mainClass"] = "org.apache.spark.deploy.PythonRunner"
        elif "mainClass" not in config:
            raise JobException("'mainClass' not set in job configuration")

    return config


def convert_git_status_to_files(files):
    # Convert GitFileStatus to list of file paths
    if isinstance(files[0], GitFileStatus):
        tmp_files = []
        for file_status in files:
            tmp_files.append(file_status.file)
        files = tmp_files

    return files


def is_interactive():
    import __main__ as main

    return not hasattr(main, "__file__")



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from datetime import date, datetime
from typing import Optional, Union

import humps
from hsfs import util
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics


class FeatureMonitoringResult:
    def __init__(
        self,
        feature_store_id: int,
        execution_id: int,
        monitoring_time: Union[int, datetime, date, str],
        config_id: int,
        feature_name: str,
        difference: Optional[float] = None,
        shift_detected: bool = False,
        detection_statistics_id: Optional[int] = None,
        reference_statistics_id: Optional[int] = None,
        empty_detection_window: bool = False,
        empty_reference_window: bool = False,
        specific_value: Optional[float] = None,
        raised_exception: bool = False,
        detection_statistics: Optional[
            Union[FeatureDescriptiveStatistics, dict]
        ] = None,
        reference_statistics: Optional[
            Union[FeatureDescriptiveStatistics, dict]
        ] = None,
        id: Optional[int] = None,
        href: Optional[str] = None,
        **kwargs,
    ):
        self._id = id
        self._href = href
        self._feature_store_id = feature_store_id
        self._execution_id = execution_id
        self._config_id = config_id
        self._feature_name = feature_name
        self._detection_statistics_id = detection_statistics_id
        self._reference_statistics_id = reference_statistics_id
        self._detection_statistics = self._parse_descriptive_statistics(
            detection_statistics
        )
        self._reference_statistics = self._parse_descriptive_statistics(
            reference_statistics
        )
        self._monitoring_time = util.convert_event_time_to_timestamp(monitoring_time)
        self._difference = difference
        self._shift_detected = shift_detected
        self._empty_detection_window = empty_detection_window
        self._empty_reference_window = empty_reference_window
        self._raised_exception = raised_exception
        self._specific_value = specific_value

    def _parse_descriptive_statistics(
        self,
        statistics: Optional[Union[FeatureDescriptiveStatistics, dict]],
    ) -> Optional[FeatureDescriptiveStatistics]:
        if statistics is None:
            return None
        return (
            statistics
            if isinstance(statistics, FeatureDescriptiveStatistics)
            else FeatureDescriptiveStatistics.from_response_json(statistics)
        )

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if json_decamelized["count"] == 0:
                return []
            return [cls(**result) for result in json_decamelized["items"]]
        else:
            return cls(**json_decamelized)

    def to_dict(self):
        the_dict = {
            "id": self._id,
            "featureStoreId": self._feature_store_id,
            "configId": self._config_id,
            "executionId": self._execution_id,
            "monitoringTime": self._monitoring_time,
            "difference": self._difference,
            "shiftDetected": self._shift_detected,
            "featureName": self._feature_name,
            "emptyDetectionWindow": self._empty_detection_window,
            "emptyReferenceWindow": self._empty_reference_window,
            "raisedException": self._raised_exception,
            "specificValue": self._specific_value,
        }

        if self._detection_statistics_id is not None:
            the_dict["detectionStatisticsId"] = self._detection_statistics_id
        if self._reference_statistics_id is not None:
            the_dict["referenceStatisticsId"] = self._reference_statistics_id
        if self._detection_statistics is not None:
            the_dict["detectionStatistics"] = self._detection_statistics.to_dict()
        if self._reference_statistics is not None:
            the_dict["referenceStatistics"] = self._reference_statistics.to_dict()

        return the_dict

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self) -> str:
        return json.dumps(humps.decamelize(self.to_dict()), indent=2)

    @property
    def id(self) -> Optional[int]:
        """Id of the feature monitoring result."""
        return self._id

    @property
    def config_id(self) -> int:
        """Id of the feature monitoring configuration containing this result."""
        return self._config_id

    @property
    def feature_store_id(self) -> int:
        """Id of the Feature Store."""
        return self._feature_store_id

    @property
    def detection_statistics_id(self) -> Optional[int]:
        """Id of the feature descriptive statistics computed on the detection window."""
        return self._detection_statistics_id

    @property
    def reference_statistics_id(self) -> Optional[int]:
        """Id of the feature descriptive statistics computed on the reference window."""
        return self._reference_statistics_id

    @property
    def detection_statistics(self) -> Optional[FeatureDescriptiveStatistics]:
        """Feature descriptive statistics computed on the detection window."""
        return self._detection_statistics

    @property
    def reference_statistics(self) -> Optional[FeatureDescriptiveStatistics]:
        """Feature descriptive statistics computed on the reference window."""
        return self._reference_statistics

    @property
    def execution_id(self) -> Optional[int]:
        """Execution id of the feature monitoring job."""
        return self._execution_id

    @property
    def monitoring_time(self) -> int:
        """Time at which this feature monitoring result was created."""
        return self._monitoring_time

    @property
    def difference(self) -> Optional[float]:
        """Difference between detection and reference values. It can be relative or absolute difference,
        depending on the statistics comparison configuration provided in `relative` parameter passed to `compare_on()`
        when enabling feature monitoring.
        """
        return self._difference

    @property
    def shift_detected(self) -> bool:
        """Whether or not shift was detected in the detection window based on the computed statistics and the threshold provided in `compare_on()`
        when enabling feature monitoring."""
        return self._shift_detected

    @property
    def feature_name(self) -> str:
        """Name of the feature being monitored."""
        return self._feature_name

    @property
    def empty_detection_window(self) -> bool:
        """Whether or not the detection window was empty in this feature monitoring run."""
        return self._empty_detection_window

    @property
    def empty_reference_window(self) -> bool:
        """Whether or not the reference window was empty in this feature monitoring run."""
        return self._empty_reference_window

    @property
    def specific_value(self) -> Optional[float]:
        """Specific value used as reference in the statistics comparison."""
        return self._specific_value



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import json
import warnings
from datetime import datetime, timezone
from typing import Literal, Optional

import humps
from hopsworks_common import client, usage, util
from hopsworks_common.client.exceptions import JobException
from hopsworks_common.core import execution_api, job_api
from hopsworks_common.engine import execution_engine
from hopsworks_common.job_schedule import JobSchedule


class Job:
    def __init__(
        self,
        id,
        name,
        creation_time,
        config,
        job_type,
        creator,
        executions=None,
        type=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        job_schedule=None,
        **kwargs,
    ):
        self._id = id
        self._name = name
        self._creation_time = creation_time
        self._config = config
        self._job_type = job_type
        self._creator = creator
        self._executions = executions
        self._href = href
        self._job_schedule = (
            JobSchedule.from_response_json(job_schedule)
            if job_schedule
            else job_schedule
        )

        self._execution_engine = execution_engine.ExecutionEngine()
        self._execution_api = execution_api.ExecutionApi()
        self._execution_engine = execution_engine.ExecutionEngine()
        self._job_api = job_api.JobApi()

    @classmethod
    def from_response_json(cls, json_dict):
        if "items" in json_dict:
            jobs = []
            for job in json_dict["items"]:
                # Job config should not be decamelized when updated
                config = job.pop("config")
                json_decamelized = humps.decamelize(job)
                json_decamelized["config"] = config
                jobs.append(cls(**json_decamelized))
            return jobs
        elif "id" not in json_dict:
            return []
        else:
            # Job config should not be decamelized when updated
            config = json_dict.pop("config")
            json_decamelized = humps.decamelize(json_dict)
            json_decamelized["config"] = config
            return cls(**json_decamelized)

    @property
    def id(self):
        """Id of the job"""
        return self._id

    @property
    def name(self):
        """Name of the job"""
        return self._name

    @property
    def creation_time(self):
        """Date of creation for the job"""
        return self._creation_time

    @property
    def config(self):
        """Configuration for the job"""
        return self._config

    @config.setter
    def config(self, config: dict):
        """Update configuration for the job"""
        self._config = config

    @property
    def job_type(self):
        """Type of the job"""
        return self._job_type

    @property
    def creator(self):
        """Creator of the job"""
        return self._creator

    @property
    def job_schedule(self):
        """Return the Job schedule"""
        return self._job_schedule

    @property
    def executions(self):
        return self._executions

    @property
    def href(self):
        return self._href

    @property
    def config(self):
        """Configuration for the job"""
        return self._config

    @usage.method_logger
    def run(self, args: str = None, await_termination: bool = True):
        """Run the job.

        Run the job, by default awaiting its completion, with the option of passing runtime arguments.

        !!! example
            ```python
            # connect to the Feature Store
            fs = ...

            # get the Feature Group instances
            fg = fs.get_or_create_feature_group(...)

            # insert in to feature group
            job, _ = fg.insert(df, write_options={"start_offline_materialization": False})

            # run job
            execution = job.run()

            # True if job executed successfully
            print(execution.success)

            # Download logs
            out_log_path, err_log_path = execution.download_logs()
            ```

        # Arguments
            args: Optional runtime arguments for the job.
            await_termination: Identifies if the client should wait for the job to complete, defaults to True.
        # Returns
            `Execution`. The execution object for the submitted run.
        """
        if self._is_materialization_running(args):
            return None
        print(f"Launching job: {self.name}")
        execution = self._execution_api._start(self, args=args)
        print(
            f"Job started successfully, you can follow the progress at \n{execution.get_url()}"
        )
        if await_termination:
            return self._execution_engine.wait_until_finished(self, execution)
        else:
            return execution

    def get_state(self):
        """Get the state of the job.

        # Returns
            `state`. Current state of the job, which can be one of the following:
            `INITIALIZING`, `INITIALIZATION_FAILED`, `FINISHED`, `RUNNING`, `ACCEPTED`,
            `FAILED`, `KILLED`, `NEW`, `NEW_SAVING`, `SUBMITTED`, `AGGREGATING_LOGS`,
            `FRAMEWORK_FAILURE`, `STARTING_APP_MASTER`, `APP_MASTER_START_FAILED`,
            `GENERATING_SECURITY_MATERIAL`, `CONVERTING_NOTEBOOK`. If no executions are found for the job,
            a warning is raised and it returns `UNDEFINED`.
        """
        last_execution = self._job_api.last_execution(self)
        if len(last_execution) != 1:
            raise JobException("No executions found for job")

        return last_execution[0].state

    def get_final_state(
        self,
    ) -> Literal[
        "UNDEFINED",
        "FINISHED",
        "FAILED",
        "KILLED",
        "FRAMEWORK_FAILURE",
        "APP_MASTER_START_FAILED",
        "INITIALIZATION_FAILED",
    ]:
        """Get the final state of the job.

        # Returns
            `final_state`. Final state of the job, which can be one of the following:
            `UNDEFINED`, `FINISHED`, `FAILED`, `KILLED`, `FRAMEWORK_FAILURE`,
            `APP_MASTER_START_FAILED`, `INITIALIZATION_FAILED`. `UNDEFINED` indicates
             that the job is still running.
        """
        last_execution = self._job_api.last_execution(self)
        if len(last_execution) != 1:
            raise JobException("No executions found for job")

        return last_execution[0].final_status

    def get_executions(self):
        """Retrieves all executions for the job ordered by submission time.

        # Returns
            `List[Execution]`
        # Raises
            `RestAPIError` in case the backend fails to retrieve executions.
        """
        return self._execution_api._get_all(self)

    @usage.method_logger
    def save(self):
        """Save the job.

        This function should be called after changing a property such as the job configuration to save it persistently.

        ```python
        job.config['appPath'] = "Resources/my_app.py"
        job.save()
        ```
        # Returns
            `Job`. The updated job object.
        """
        return self._job_api._update_job(self.name, self.config)

    @usage.method_logger
    def delete(self):
        """Delete the job
        !!! danger "Potentially dangerous operation"
            This operation deletes the job and all executions.
        # Raises
            `RestAPIError`.
        """
        self._job_api._delete(self)

    def _is_materialization_running(self, args: str) -> bool:
        if self.name.endswith("offline_fg_materialization") or self.name.endswith(
            "offline_fg_backfill"
        ):
            try:
                should_abort = self.get_final_state() == "UNDEFINED"
            except JobException as e:
                if "No executions found for job" in str(e):
                    should_abort = False
                else:
                    raise e
            if should_abort:
                warnings.warn(
                    "Materialization job is already running, aborting new execution."
                    "Please wait for the current execution to finish before triggering a new one."
                    "You can check the status of the current execution using `fg.materialization_job.get_state()`."
                    "or `fg.materialization_job.get_final_state()` or check it out in the Hopsworks UI."
                    f"at {self.get_url()}.\n"
                    f"Use fg.materialization_job.run(args={args}) to trigger the materialization job again.",
                    stacklevel=2,
                )
                return True
        return False

    def _wait_for_job(self, await_termination=True, timeout: Optional[float] = None):
        # If the user passed the wait_for_job option consider it,
        # otherwise use the default True
        if await_termination:
            executions = self._job_api.last_execution(self)
            if len(executions) > 0:
                execution = executions[0]
            else:
                return
            self._execution_engine.wait_until_finished(
                job=self, execution=execution, timeout=timeout
            )

    def schedule(
        self,
        cron_expression: str,
        start_time: datetime = None,
        end_time: datetime = None,
    ):
        """Schedule the execution of the job.

        If a schedule for this job already exists, the method updates it.

        ```python
        # Schedule the job
        job.schedule(
            cron_expression="0 */5 * ? * * *",
            start_time=datetime.datetime.now(tz=timezone.utc)
        )

        # Retrieve the next execution time
        print(job.job_schedule.next_execution_date_time)
        ```

        # Arguments
            cron_expression: The quartz cron expression
            start_time: The schedule start time in UTC. If None, the current time is used. The start_time can be a value in the past.
            end_time: The schedule end time in UTC. If None, the schedule will continue running indefinitely. The end_time can be a value in the past.
        # Returns
            `JobSchedule`. The schedule of the job
        """
        job_schedule = JobSchedule(
            id=self._job_schedule.id if self._job_schedule else None,
            start_date_time=start_time if start_time else datetime.now(tz=timezone.utc),
            cron_expression=cron_expression,
            end_time=end_time,
            enabled=True,
        )
        self._job_schedule = self._job_api._schedule_job(
            self._name, job_schedule.to_dict()
        )
        return self._job_schedule

    @usage.method_logger
    def unschedule(self):
        """Unschedule the exceution of a Job"""
        self._job_api._delete_schedule_job(self._name)
        self._job_schedule = None

    @usage.method_logger
    def resume_schedule(self):
        """Resumes the schedule of a Job execution"""
        if self._job_schedule is None:
            raise JobException("No schedule found for job")

        job_schedule = JobSchedule(
            id=self._job_schedule.id,
            start_date_time=self._job_schedule.start_date_time,
            cron_expression=self._job_schedule.cron_expression,
            end_time=self._job_schedule.end_date_time,
            enabled=False,
        )
        return self._update_schedule(job_schedule)

    @usage.method_logger
    def pause_schedule(self):
        """Pauses the schedule of a Job execution"""
        if self._job_schedule is None:
            raise JobException("No schedule found for job")

        job_schedule = JobSchedule(
            id=self._job_schedule.id,
            start_date_time=self._job_schedule.start_date_time,
            cron_expression=self._job_schedule.cron_expression,
            end_time=self._job_schedule.end_date_time,
            enabled=True,
        )
        return self._update_schedule(job_schedule)

    def _update_schedule(self, job_schedule):
        self._job_schedule = self._job_api.create_or_update_schedule_job(
            self._name, job_schedule.to_dict()
        )
        return self._job_schedule

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"Job({self._name!r}, {self._job_type!r})"

    def get_url(self):
        _client = client.get_instance()
        path = "/p/" + str(_client._project_id) + "/jobs/named/" + self.name
        return util.get_hostname_replaced_url(path)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import logging
from typing import List, Union

from hopsworks_common import (
    client,
    git_commit,
    git_file_status,
    git_op_execution,
    git_repo,
    usage,
    util,
)
from hopsworks_common.client.exceptions import GitException
from hopsworks_common.core import git_provider_api
from hopsworks_common.engine import git_engine
from hopsworks_common.git_file_status import GitFileStatus


class GitApi:
    def __init__(self):
        self._git_engine = git_engine.GitEngine()
        self._git_provider_api = git_provider_api.GitProviderApi()
        self._log = logging.getLogger(__name__)

    @usage.method_logger
    def clone(self, url: str, path: str, provider: str = None, branch: str = None):
        """Clone a new Git Repo in to Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        git_api = project.get_git_api()

        git_repo = git_api.clone("https://github.com/logicalclocks/hops-examples.git", "Resources", "GitHub")

        ```
        # Arguments
            url: Url to the git repository
            path: Path in Hopsworks Filesystem to clone the repo to
            provider: The git provider where the repo is currently hosted. Valid values are "GitHub", "GitLab" and "BitBucket".
            branch: Optional branch to clone, defaults to configured main branch
        # Returns
            `GitRepo`: Git repository object
        # Raises
            `RestAPIError`: If unable to clone the git repository.
        """

        _client = client.get_instance()

        # Support absolute and relative path to dataset
        path = util.convert_to_abs(path, _client._project_name)

        if provider is None:
            provider = self._git_provider_api._get_default_configured_provider()

        path_params = ["project", _client._project_id, "git", "clone"]

        clone_config = {
            "url": url,
            "path": path,
            "provider": provider,
            "branch": branch,
        }
        query_params = {"expand": ["repository", "user"]}

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                data=json.dumps(clone_config),
                query_params=query_params,
            )
        )
        print(
            "Git clone operation running, explore it at "
            + util.get_hostname_replaced_url(
                "/p/" + str(_client._project_id) + "/settings/git"
            )
        )
        git_op = self._git_engine.execute_op_blocking(git_op, "CLONE")
        created_repo = self.get_repo(git_op.repository.name, git_op.repository.path)
        return created_repo

    @usage.method_logger
    def get_repos(self):
        """Get the existing Git repositories

        # Returns
            `List[GitRepo]`: List of git repository objects
        # Raises
            `RestAPIError`: If unable to get the repositories
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
        ]
        query_params = {"expand": "creator"}
        return git_repo.GitRepo.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

    @usage.method_logger
    def get_providers(self):
        """Get the configured Git providers

        # Returns
            `List[GitProvider]`: List of git provider objects
        # Raises
            `RestAPIError`: If unable to get the git providers
        """
        return self._git_provider_api._get_providers()

    @usage.method_logger
    def get_provider(self, provider: str):
        """Get the configured Git provider

        # Arguments
            provider: Name of git provider. Valid values are "GitHub", "GitLab" and "BitBucket".
        # Returns
            `GitProvider`: The git provider
        # Raises
            `RestAPIError`: If unable to get the git provider
        """
        return self._git_provider_api._get_provider(provider)

    @usage.method_logger
    def set_provider(self, provider: str, username: str, token: str):
        """Configure a Git provider

        ```python

        import hopsworks

        project = hopsworks.login()

        git_api = project.get_git_api()

        git_api.set_provider("GitHub", "my_user", "my_token")

        ```
        # Arguments
            provider: Name of git provider. Valid values are "GitHub", "GitLab" and "BitBucket".
            username: Username for the git provider service
            token: Token to set for the git provider service
        # Raises
            `RestAPIError`: If unable to configure the git provider
        """
        self._git_provider_api._set_provider(provider, username, token)

    @usage.method_logger
    def get_repo(self, name: str, path: str = None):
        """Get the cloned Git repository

        # Arguments
            name: Name of git repository
            path: Optional path to specify if multiple git repos with the same name exists in the project
        # Returns
            `GitRepo`: The git repository
        # Raises
            `RestAPIError`: If unable to get the git repository
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
        ]
        query_params = {"expand": "creator"}

        repos = git_repo.GitRepo.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

        if path is not None:
            path = util.convert_to_abs(path, _client._project_name)

        filtered_repos = []
        for repository in repos:
            if repository.name == name:
                if path is None:
                    filtered_repos.append(repository)
                elif repository.path == path:
                    filtered_repos.append(repository)

        if len(filtered_repos) == 1:
            return filtered_repos[0]
        elif len(filtered_repos) > 1:
            raise GitException(
                "Multiple repositories found matching name {}. Please specify the repository by setting the path keyword, for example path='Resources/{}'.".format(
                    name, name
                )
            )
        else:
            raise GitException("No git repository found matching name {}".format(name))

    def _delete_repo(self, repo_id):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
        ]
        _client._send_request("DELETE", path_params)

    def _create(self, repo_id, branch: str, checkout=False):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
            "branch",
        ]

        query_params = {"branchName": branch, "expand": "repository"}

        if checkout:
            query_params["action"] = "CREATE_CHECKOUT"
        else:
            query_params["action"] = "CREATE"

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, query_params=query_params
            )
        )
        _ = self._git_engine.execute_op_blocking(git_op, query_params["action"])

    def _delete(self, repo_id, branch: str):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
            "branch",
        ]

        query_params = {
            "action": "DELETE",
            "branchName": branch,
            "expand": "repository",
        }

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, query_params=query_params
            )
        )
        _ = self._git_engine.execute_op_blocking(git_op, query_params["action"])

    def _checkout(
        self, repo_id, branch: str = None, commit: str = None, force: bool = False
    ):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
            "branch",
        ]

        query_params = {"branchName": branch, "commit": commit, "expand": "repository"}

        if force:
            query_params["action"] = "CHECKOUT_FORCE"
        else:
            query_params["action"] = "CHECKOUT"

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, query_params=query_params
            )
        )
        _ = self._git_engine.execute_op_blocking(git_op, query_params["action"])

    def _status(self, repo_id):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
        ]

        query_params = {"action": "STATUS", "expand": ["repository", "user"]}

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                query_params=query_params,
            )
        )
        git_op = self._git_engine.execute_op_blocking(git_op, query_params["action"])

        status_dict = json.loads(git_op.command_result_message)
        file_status = None
        if status_dict is not None and isinstance(status_dict["status"], list):
            file_status = []
            for status in status_dict["status"]:
                file_status.append(
                    git_file_status.GitFileStatus.from_response_json(status)
                )
        else:
            self._log.info("Nothing to commit, working tree clean")

        return file_status

    def _commit(self, repo_id, message: str, all=False, files=None):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
        ]

        query_params = {"action": "COMMIT", "expand": ["repository", "user"]}
        commit_config = {
            "type": "commitCommandConfiguration",
            "message": message,
            "all": all,
            "files": files,
        }

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                query_params=query_params,
                data=json.dumps(commit_config),
            )
        )
        _ = self._git_engine.execute_op_blocking(git_op, query_params["action"])

    def _push(self, repo_id, branch: str, force: bool = False, remote: str = None):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
        ]

        query_params = {"action": "PUSH", "expand": ["repository", "user"]}
        push_config = {
            "type": "pushCommandConfiguration",
            "remoteName": remote,
            "force": force,
            "branchName": branch,
        }

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                query_params=query_params,
                data=json.dumps(push_config),
            )
        )
        _ = self._git_engine.execute_op_blocking(git_op, query_params["action"])

    def _pull(self, repo_id, branch: str, force: bool = False, remote: str = None):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
        ]

        query_params = {"action": "PULL", "expand": ["repository", "user"]}
        push_config = {
            "type": "pullCommandConfiguration",
            "remoteName": remote,
            "force": force,
            "branchName": branch,
        }

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                query_params=query_params,
                data=json.dumps(push_config),
            )
        )
        _ = self._git_engine.execute_op_blocking(git_op, query_params["action"])

    def _checkout_files(self, repo_id, files: Union[List[str], List[GitFileStatus]]):
        files = util.convert_git_status_to_files(files)

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
            "file",
        ]

        query_params = {"expand": ["repository", "user"]}

        headers = {"content-type": "application/json"}
        git_op = git_op_execution.GitOpExecution.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                query_params=query_params,
                data=json.dumps({"files": files}),
            )
        )
        _ = self._git_engine.execute_op_blocking(git_op, "CHECKOUT_FILES")

    def _get_commits(self, repo_id, branch: str):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
            "branch",
            branch,
            "commit",
        ]

        headers = {"content-type": "application/json"}
        return git_commit.GitCommit.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import ge_validation_result


class TestValidationResult:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_validation_result"]["get"]["response"]

        # Act
        vr = ge_validation_result.ValidationResult.from_response_json(json)

        # Assert
        assert vr.id == 11
        assert vr.success is True
        assert vr._observed_value == "test_observed_value"
        assert vr._expectation_id == 22
        assert vr._validation_report_id == 33
        assert vr.result == {"result_key": "result_value"}
        assert vr.meta == {"meta_key": "meta_value"}
        assert vr.exception_info == {"exception_info_key": "exception_info_value"}
        assert vr.expectation_config == {
            "expectation_config_key": "expectation_config_value"
        }

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_validation_result"]["get_basic_info"]["response"]

        # Act
        vr = ge_validation_result.ValidationResult.from_response_json(json)

        # Assert
        assert vr.id is None
        assert vr.success is True
        assert vr._observed_value is None
        assert vr._expectation_id is None
        assert vr._validation_report_id is None
        assert vr.result == {"result_key": "result_value"}
        assert vr.meta == {"meta_key": "meta_value"}
        assert vr.exception_info == {"exception_info_key": "exception_info_value"}
        assert vr.expectation_config == {
            "expectation_config_key": "expectation_config_value"
        }

    def test_from_response_json_list(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_validation_result"]["get_list"]["response"]

        # Act
        vr_list = ge_validation_result.ValidationResult.from_response_json(json)

        # Assert
        assert len(vr_list) == 1
        vr = vr_list[0]
        assert vr.id == 11
        assert vr.success is True
        assert vr._observed_value == "test_observed_value"
        assert vr._expectation_id == 22
        assert vr._validation_report_id == 33
        assert vr.result == {"result_key": "result_value"}
        assert vr.meta == {"meta_key": "meta_value"}
        assert vr.exception_info == {"exception_info_key": "exception_info_value"}
        assert vr.expectation_config == {
            "expectation_config_key": "expectation_config_value"
        }

    def test_from_response_json_list_empty(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_validation_result"]["get_list_empty"]["response"]

        # Act
        vr_list = ge_validation_result.ValidationResult.from_response_json(json)

        # Assert
        assert len(vr_list) == 0

    def test_timezone_support_validation_time(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["ge_validation_result"]["get"]["response"]
        json["validation_time"] = "2023-02-15T09:20:03.000414-05"

        # Act
        vr = ge_validation_result.ValidationResult.from_response_json(json)

        # Assert
        assert vr.id == 11
        assert vr.validation_time == 1676470803000



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Optional

from hopsworks_common import client
from hsfs import expectation_suite as es
from hsfs.core.variable_api import VariableApi


class ExpectationSuiteApi:
    def __init__(self, feature_store_id: int, feature_group_id: int) -> None:
        """Expectation Suite endpoints for the Feature Group resource.

        :param feature_store_id: id of the respective Feature Store
        :type feature_store_id: int
        :param feature_group_id: id of the respective Feature Group
        :type feature_group_id: int
        """
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._variable_api = VariableApi()

    def create(self, expectation_suite: "es.ExpectationSuite") -> "es.ExpectationSuite":
        """Create an expectation suite attached to a Feature Group.

        :param expectation_suite: expectation suite object to be created for a Feature Group
        :type expectation_suite: `ExpectationSuite`
        :return: the created expectation suite
        :rtype: ExpectationSuite
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
        ]

        major, minor = self._variable_api.parse_major_and_minor(
            self._variable_api.get_version("hopsworks")
        )
        method = "POST"
        if major == "3" and minor == "0":
            method = "PUT"

        headers = {"content-type": "application/json"}
        payload = expectation_suite.json()
        return es.ExpectationSuite.from_response_json(
            _client._send_request(method, path_params, headers=headers, data=payload)
        )

    def update(self, expectation_suite: "es.ExpectationSuite") -> "es.ExpectationSuite":
        """Update an expectation suite attached to a Feature Group.

        :param expectation_suite: expectation suite object to be created for a Feature Group
        :type expectation_suite: `ExpectationSuite`
        :return: the updated expectation suite
        :rtype: ExpectationSuite
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            expectation_suite.id,
        ]

        headers = {"content-type": "application/json"}
        payload = expectation_suite.json()

        major, minor = self._variable_api.parse_major_and_minor(
            self._variable_api.get_version("hopsworks")
        )
        method = "PUT"
        if major == "3" and minor == "0":
            method = "POST"
            del path_params[-1]

        return es.ExpectationSuite.from_response_json(
            _client._send_request(method, path_params, headers=headers, data=payload)
        )

    def update_metadata(
        self, expectation_suite: "es.ExpectationSuite"
    ) -> "es.ExpectationSuite":
        """Update the metadata of an expectation suite attached to a Feature Group.

        :param expectation_suite: expectation suite object to be updated
        :type expectation_suite: `ExpectationSuite`
        :return: the expectation suite with updated metadata.
        :rtype: ExpectationSuite
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            expectation_suite.id,
            "metadata",
        ]

        headers = {"content-type": "application/json"}
        payload = expectation_suite.json()

        major, minor = self._variable_api.parse_major_and_minor(
            self._variable_api.get_version("hopsworks")
        )
        method = "PUT"
        if major == "3" and minor == "0":
            method = "POST"
            del path_params[-1]
            del path_params[-1]

        return es.ExpectationSuite.from_response_json(
            _client._send_request(method, path_params, headers=headers, data=payload)
        )

    def delete(self, expectation_suite_id: int) -> None:
        """Delete the expectation suite attached to a Feature Group."""
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
            expectation_suite_id,
        ]

        major, minor = self._variable_api.parse_major_and_minor(
            self._variable_api.get_version("hopsworks")
        )
        if major == "3" and minor == "0":
            del path_params[-1]

        _client._send_request("DELETE", path_params)

    def get(self) -> Optional["es.ExpectationSuite"]:
        """Get the expectation suite attached to a Feature Group.

        :return: fetched expectation suite attached to the FeatureG Group
        :rtype: ExpectationSuite || None
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "expectationsuite",
        ]

        return es.ExpectationSuite.from_response_json(
            _client._send_request("GET", path_params)
        )



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

from typing import Literal, Optional, Union

from hopsworks_common.client import external, hopsworks, istio
from hopsworks_common.constants import HOSTS


_client: Union[hopsworks.Client, external.Client, None] = None


def init(
    client_type: Union[Literal["hopsworks"], Literal["external"]],
    host: Optional[str] = None,
    port: Optional[int] = None,
    project: Optional[str] = None,
    engine: Optional[str] = None,
    hostname_verification: Optional[bool] = None,
    trust_store_path: Optional[str] = None,
    cert_folder: Optional[str] = None,
    api_key_file: Optional[str] = None,
    api_key_value: Optional[str] = None,
) -> None:
    global _client
    if not _client:
        if client_type == "hopsworks":
            _client = hopsworks.Client(hostname_verification)
        elif client_type == "external":
            _client = external.Client(
                host,
                port,
                project,
                engine,
                hostname_verification,
                trust_store_path,
                cert_folder,
                api_key_file,
                api_key_value,
            )
    elif _client._is_external() and not _client._project_name:
        _client.provide_project(project)


def get_instance() -> Union[hopsworks.Client, external.Client]:
    global _client
    if not _client:
        raise Exception("Couldn't find client. Try reconnecting to Hopsworks.")
    return _client


def stop() -> None:
    global _client
    if _client:
        _client._close()
    _client = None
    if istio._client:
        istio._client._close()
    istio._client = None


def is_saas_connection() -> bool:
    return get_instance()._host == HOSTS.APP_HOST


_kserve_installed = None


def set_kserve_installed(kserve_installed):
    global _kserve_installed
    _kserve_installed = kserve_installed


def is_kserve_installed() -> bool:
    global _kserve_installed
    return _kserve_installed


_serving_resource_limits = None


def set_serving_resource_limits(max_resources):
    global _serving_resource_limits
    _serving_resource_limits = max_resources


def get_serving_resource_limits():
    global _serving_resource_limits
    return _serving_resource_limits


_serving_num_instances_limits = None


def set_serving_num_instances_limits(num_instances_range):
    global _serving_num_instances_limits
    _serving_num_instances_limits = num_instances_range


def get_serving_num_instances_limits():
    global _serving_num_instances_limits
    return _serving_num_instances_limits


def is_scale_to_zero_required():
    # scale-to-zero is required for KServe deployments if the Hopsworks variable `kube_serving_min_num_instances`
    # is set to 0. Other possible values are -1 (unlimited num instances) or >1 num instances.
    return get_serving_num_instances_limits()[0] == 0


_knative_domain = None


def get_knative_domain():
    global _knative_domain
    return _knative_domain


def set_knative_domain(knative_domain):
    global _knative_domain
    _knative_domain = knative_domain


_connection = None


def get_connection():
    return _connection


def set_connection(connection):
    global _connection
    _connection = connection


def _is_external():
    global _client
    if _client is None:
        raise ConnectionError("Hopsworks Client not initialized.")
    return _client._is_external()



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import copy
import logging
import math
import os
import shutil
import time
from concurrent.futures import ThreadPoolExecutor, wait

from hopsworks_common import client, usage, util
from hopsworks_common.client.exceptions import DatasetException, RestAPIError
from hopsworks_common.core import inode
from tqdm.auto import tqdm


class Chunk:
    def __init__(self, content, number, status):
        self.content = content
        self.number = number
        self.status = status
        self.retries = 0


class DatasetApi:
    def __init__(self):
        self._log = logging.getLogger(__name__)

    DEFAULT_FLOW_CHUNK_SIZE = 1048576
    FLOW_PERMANENT_ERRORS = [404, 413, 415, 500, 501]

    @usage.method_logger
    def download(self, path: str, local_path: str = None, overwrite: bool = False):
        """Download file from Hopsworks Filesystem to the current working directory.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        downloaded_file_path = dataset_api.download("Resources/my_local_file.txt")

        ```
        # Arguments
            path: path in Hopsworks filesystem to the file
            local_path: path where to download the file in the local filesystem
            overwrite: overwrite local file if exists
        # Returns
            `str`: Path to downloaded file
        # Raises
            `RestAPIError`: If unable to download the file
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "dataset",
            "download",
            "with_auth",
            path,
        ]
        query_params = {"type": "DATASET"}

        # Build the path to download the file on the local fs and return to the user, it should be absolute for consistency
        # Download in CWD if local_path not specified
        if local_path is None:
            local_path = os.path.join(os.getcwd(), os.path.basename(path))
        # If local_path specified, ensure it is absolute
        else:
            if os.path.isabs(local_path):
                local_path = os.path.join(local_path, os.path.basename(path))
            else:
                local_path = os.path.join(
                    os.getcwd(), local_path, os.path.basename(path)
                )

        if os.path.exists(local_path):
            if overwrite:
                if os.path.isfile:
                    os.remove(local_path)
                else:
                    shutil.rmtree(local_path)
            else:
                raise IOError(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        local_path
                    )
                )

        file_size = int(self._get(path)["attributes"]["size"])
        with _client._send_request(
            "GET", path_params, query_params=query_params, stream=True
        ) as response:
            with open(local_path, "wb") as f:
                pbar = None
                try:
                    pbar = tqdm(
                        total=file_size,
                        bar_format="{desc}: {percentage:.3f}%|{bar}| {n_fmt}/{total_fmt} elapsed<{elapsed} remaining<{remaining}",
                        desc="Downloading",
                    )
                except Exception:
                    self._log.exception("Failed to initialize progress bar.")
                    self._log.info("Starting download")

                for chunk in response.iter_content(
                    chunk_size=self.DEFAULT_FLOW_CHUNK_SIZE
                ):
                    f.write(chunk)

                    if pbar is not None:
                        pbar.update(len(chunk))

                if pbar is not None:
                    pbar.close()
                else:
                    self._log.info("Download finished")

        return local_path

    @usage.method_logger
    def upload(
        self,
        local_path: str,
        upload_path: str,
        overwrite: bool = False,
        chunk_size=DEFAULT_FLOW_CHUNK_SIZE,
        simultaneous_uploads=3,
        max_chunk_retries=1,
        chunk_retry_interval=1,
    ):
        """Upload a file to the Hopsworks filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        uploaded_file_path = dataset_api.upload("my_local_file.txt", "Resources")

        ```
        # Arguments
            local_path: local path to file to upload
            upload_path: path to directory where to upload the file in Hopsworks Filesystem
            overwrite: overwrite file if exists
            chunk_size: upload chunk size in bytes. Default 1048576 bytes
            simultaneous_uploads: number of simultaneous chunks to upload. Default 3
            max_chunk_retries: maximum retry for a chunk. Default is 1
            chunk_retry_interval: chunk retry interval in seconds. Default is 1sec
        # Returns
            `str`: Path to uploaded file
        # Raises
            `RestAPIError`: If unable to upload the file
        """
        # local path could be absolute or relative,
        if not os.path.isabs(local_path) and os.path.exists(
            os.path.join(os.getcwd(), local_path)
        ):
            local_path = os.path.join(os.getcwd(), local_path)

        file_size = os.path.getsize(local_path)

        _, file_name = os.path.split(local_path)

        destination_path = upload_path + "/" + file_name

        if self.exists(destination_path):
            if overwrite:
                self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        local_path
                    )
                )

        num_chunks = math.ceil(file_size / chunk_size)

        base_params = self._get_flow_base_params(
            file_name, num_chunks, file_size, chunk_size
        )

        chunk_number = 1
        with open(local_path, "rb") as f:
            pbar = None
            try:
                pbar = tqdm(
                    total=file_size,
                    bar_format="{desc}: {percentage:.3f}%|{bar}| {n_fmt}/{total_fmt} elapsed<{elapsed} remaining<{remaining}",
                    desc="Uploading",
                )
            except Exception:
                self._log.exception("Failed to initialize progress bar.")
                self._log.info("Starting upload")
            with ThreadPoolExecutor(simultaneous_uploads) as executor:
                while True:
                    chunks = []
                    for _ in range(simultaneous_uploads):
                        chunk = f.read(chunk_size)
                        if not chunk:
                            break
                        chunks.append(Chunk(chunk, chunk_number, "pending"))
                        chunk_number += 1

                    if len(chunks) == 0:
                        break

                    # upload each chunk and update pbar
                    futures = [
                        executor.submit(
                            self._upload_chunk,
                            base_params,
                            upload_path,
                            file_name,
                            chunk,
                            pbar,
                            max_chunk_retries,
                            chunk_retry_interval,
                        )
                        for chunk in chunks
                    ]
                    # wait for all upload tasks to complete
                    _, _ = wait(futures)
                    try:
                        _ = [future.result() for future in futures]
                    except Exception as e:
                        if pbar is not None:
                            pbar.close()
                        raise e

            if pbar is not None:
                pbar.close()
            else:
                self._log.info("Upload finished")

        return upload_path + "/" + os.path.basename(local_path)

    def _upload_chunk(
        self,
        base_params,
        upload_path,
        file_name,
        chunk: Chunk,
        pbar,
        max_chunk_retries,
        chunk_retry_interval,
    ):
        query_params = copy.copy(base_params)
        query_params["flowCurrentChunkSize"] = len(chunk.content)
        query_params["flowChunkNumber"] = chunk.number

        chunk.status = "uploading"
        while True:
            try:
                self._upload_request(
                    query_params, upload_path, file_name, chunk.content
                )
                break
            except RestAPIError as re:
                chunk.retries += 1
                if (
                    re.response.status_code in DatasetApi.FLOW_PERMANENT_ERRORS
                    or chunk.retries > max_chunk_retries
                ):
                    chunk.status = "failed"
                    raise re
                time.sleep(chunk_retry_interval)
                continue

        chunk.status = "uploaded"

        if pbar is not None:
            pbar.update(query_params["flowCurrentChunkSize"])

    def _get_flow_base_params(self, file_name, num_chunks, size, chunk_size):
        return {
            "templateId": -1,
            "flowChunkSize": chunk_size,
            "flowTotalSize": size,
            "flowIdentifier": str(size) + "_" + file_name,
            "flowFilename": file_name,
            "flowRelativePath": file_name,
            "flowTotalChunks": num_chunks,
        }

    def _upload_request(self, params, path, file_name, chunk):
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", "upload", path]

        # Flow configuration params are sent as form data
        _client._send_request(
            "POST", path_params, data=params, files={"file": (file_name, chunk)}
        )

    def _get(self, path: str):
        """Get dataset.

        :param path: path to check
        :type path: str
        :return: dataset metadata
        :rtype: dict
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", path]
        headers = {"content-type": "application/json"}
        return _client._send_request("GET", path_params, headers=headers)

    def exists(self, path: str):
        """Check if a file exists in the Hopsworks Filesystem.

        # Arguments
            path: path to check
        # Returns
            `bool`: True if exists, otherwise False
        # Raises
            `RestAPIError`: If unable to check existence for the path
        """
        try:
            self._get(path)
            return True
        except RestAPIError:
            return False

    @usage.method_logger
    def remove(self, path: str):
        """Remove a path in the Hopsworks Filesystem.

        # Arguments
            path: path to remove
        # Raises
            `RestAPIError`: If unable to remove the path
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", path]
        _client._send_request("DELETE", path_params)

    @usage.method_logger
    def mkdir(self, path: str):
        """Create a directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.mkdir("Resources/my_dir")

        ```
        # Arguments
            path: path to directory
        # Returns
            `str`: Path to created directory
        # Raises
            `RestAPIError`: If unable to create the directory
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", path]
        query_params = {
            "action": "create",
            "searchable": "true",
            "generate_readme": "false",
            "type": "DATASET",
        }
        headers = {"content-type": "application/json"}
        return _client._send_request(
            "POST", path_params, headers=headers, query_params=query_params
        )["attributes"]["path"]

    @usage.method_logger
    def copy(self, source_path: str, destination_path: str, overwrite: bool = False):
        """Copy a file or directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.copy("Resources/myfile.txt", "Logs/myfile.txt")

        ```
        # Arguments
            source_path: the source path to copy
            destination_path: the destination path
            overwrite: overwrite destination if exists
        # Raises
            `RestAPIError`: If unable to perform the copy
        """
        if self.exists(destination_path):
            if overwrite:
                self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        destination_path
                    )
                )

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", source_path]
        query_params = {
            "action": "copy",
            "destination_path": destination_path,
        }
        _client._send_request("POST", path_params, query_params=query_params)

    @usage.method_logger
    def move(self, source_path: str, destination_path: str, overwrite: bool = False):
        """Move a file or directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.move("Resources/myfile.txt", "Logs/myfile.txt")

        ```
        # Arguments
            source_path: the source path to move
            destination_path: the destination path
            overwrite: overwrite destination if exists
        # Raises
            `RestAPIError`: If unable to perform the move
        """

        if self.exists(destination_path):
            if overwrite:
                self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        destination_path
                    )
                )

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "dataset", source_path]
        query_params = {
            "action": "move",
            "destination_path": destination_path,
        }
        _client._send_request("POST", path_params, query_params=query_params)

    @usage.method_logger
    def upload_feature_group(self, feature_group, path, dataframe):
        # Convert the dataframe into PARQUET for upload
        df_parquet = dataframe.to_parquet(index=False)
        parquet_length = len(df_parquet)
        num_chunks = math.ceil(parquet_length / self.DEFAULT_FLOW_CHUNK_SIZE)

        base_params = self._get_flow_base_params(
            feature_group, num_chunks, parquet_length, self.DEFAULT_FLOW_CHUNK_SIZE
        )

        chunk_number = 1
        for i in range(0, parquet_length, self.DEFAULT_FLOW_CHUNK_SIZE):
            query_params = base_params
            query_params["flowCurrentChunkSize"] = len(
                df_parquet[i : i + self.DEFAULT_FLOW_CHUNK_SIZE]
            )
            query_params["flowChunkNumber"] = chunk_number

            self._upload_request(
                query_params,
                path,
                util.feature_group_name(feature_group),
                df_parquet[i : i + self.DEFAULT_FLOW_CHUNK_SIZE],
            )

            chunk_number += 1

    @usage.method_logger
    def list_files(self, path, offset, limit):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "dataset",
            path[(path.index("/", 10) + 1) :],
        ]
        query_params = {
            "action": "listing",
            "offset": offset,
            "limit": limit,
            "sort_by": "ID:asc",
        }

        inode_lst = _client._send_request("GET", path_params, query_params)

        return inode_lst["count"], inode.Inode.from_response_json(inode_lst)

    @usage.method_logger
    def read_content(self, path: str, dataset_type: str = "DATASET"):
        _client = client.get_instance()

        path_params = [
            "project",
            _client._project_id,
            "dataset",
            "download",
            "with_auth",
            path[1:],
        ]

        query_params = {
            "type": dataset_type,
        }

        return _client._send_request("GET", path_params, query_params, stream=True)



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsml import model_schema


class TestModelSchema:
    # constructor

    def test_constructor(self):
        # Act
        msch = model_schema.ModelSchema(input_schema="1234", output_schema="4321")

        # Assert
        assert msch.input_schema == "1234"
        assert msch.output_schema == "4321"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json

import humps
from hopsworks_common import usage, util
from hopsworks_common.core import kafka_api


class KafkaSchema:
    def __init__(
        self,
        id=None,
        subject=None,
        version=None,
        schema=None,
        type=None,
        **kwargs,
    ):
        self._id = id
        self._subject = subject
        self._version = version
        self._schema = schema

        self._kafka_api = kafka_api.KafkaApi()

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" not in json_decamelized:
            return cls(**json_decamelized)
        elif json_decamelized["count"] == 0:
            return []
        else:
            return [cls(**kafka_topic) for kafka_topic in json_decamelized["items"]]

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**json_decamelized)
        return self

    @property
    def id(self):
        """Id of the kafka schema"""
        return self._id

    @property
    def subject(self):
        """Name of the subject for the schema"""
        return self._subject

    @property
    def version(self):
        """Version of the schema"""
        return self._version

    @property
    def schema(self):
        """Schema definition"""
        return self._schema

    @usage.method_logger
    def delete(self):
        """Delete the schema
        !!! danger "Potentially dangerous operation"
            This operation deletes the schema.
        # Raises
            `RestAPIError`.
        """
        self._kafka_api._delete_subject_version(self.subject, self.version)

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"KafkaSchema({self._subject!r}, {self._version!r})"



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from hopsworks_common import client
from hsfs import feature_group, feature_view, statistics, training_dataset
from hsfs.core import job


class StatisticsApi:
    def __init__(self, feature_store_id: int, entity_type: str) -> None:
        """Statistics endpoint for `trainingdatasets` and `featuregroups` resource.

        :param feature_store_id: id of the respective featurestore
        :type feature_store_id: int
        :param entity_type: "trainingdatasets" or "featuregroups" or "featureview"
        :type entity_type: str
        """
        self._feature_store_id = feature_store_id
        self._entity_type = entity_type  # TODO: Support FV

    def post(
        self,
        metadata_instance: Union[
            feature_view.FeatureView,
            training_dataset.TrainingDataset,
            feature_group.FeatureGroup,
        ],
        stats: statistics.Statistics,
        training_dataset_version: Optional[int],
    ) -> Optional[statistics.Statistics]:
        _client = client.get_instance()
        path_params = self.get_path(metadata_instance, training_dataset_version)

        headers = {"content-type": "application/json"}
        stats = statistics.Statistics.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, data=stats.json()
            )
        )
        return self._extract_single_stats(stats)

    def get(
        self,
        metadata_instance: Union[
            feature_view.FeatureView,
            training_dataset.TrainingDataset,
            feature_group.FeatureGroup,
        ],
        computation_time: Optional[int] = None,
        start_commit_time: Optional[int] = None,
        end_commit_time: Optional[int] = None,
        feature_names: List[str] = None,
        row_percentage: Optional[float] = None,
        before_transformation: Optional[bool] = None,
        training_dataset_version: Optional[int] = None,
    ) -> Optional[statistics.Statistics]:
        """Get single statistics of an entity.

        :param metadata_instance: metadata object of the instance to get statistics of
        :type metadata_instance: TrainingDataset, FeatureGroup, FeatureView
        :param computation_time: Time at which statistics where computed
        :type computation_time: int
        :param start_commit_time: Window start commit time
        :type start_time: int
        :param end_commit_time: Window end commit time
        :type end_time: int
        :param feature_names: List of feature names of which statistics are retrieved
        :type feature_names: List[str]
        :param row_percentage: Percentage of feature values used during statistics computation
        :type row_percentage: float
        :param before_transformation: Whether the statistics were computed before transformations or not
        :type before_transformation: bool
        :param training_dataset_version: Version of the training dataset on which statistics were computed
        :type training_dataset_version: int
        """
        # get statistics by entity + filters + sorts, including the feature descriptive statistics
        _client = client.get_instance()
        path_params = self.get_path(metadata_instance, training_dataset_version)

        # single statistics
        offset, limit = 0, 1

        headers = {"content-type": "application/json"}
        query_params = self._build_get_query_params(
            computation_time=computation_time,
            start_commit_time=start_commit_time,
            end_commit_time=end_commit_time,
            filter_eq_times=True,
            feature_names=feature_names,
            row_percentage=row_percentage,
            before_transformation=before_transformation,
            training_dataset_version=training_dataset_version,
            # retrieve only one entity statistics, including the feature descriptive statistics
            offset=offset,
            limit=limit,
            with_content=True,
        )

        # response is either a single item or not found exception
        stats = statistics.Statistics.from_response_json(
            _client._send_request("GET", path_params, query_params, headers=headers)
        )
        return self._extract_single_stats(stats)

    def get_all(
        self,
        metadata_instance: Union[
            feature_view.FeatureView,
            training_dataset.TrainingDataset,
            feature_group.FeatureGroup,
        ],
        computation_time: Optional[int] = None,
        start_commit_time: Optional[int] = None,
        end_commit_time: Optional[int] = None,
        feature_names: Optional[List[str]] = None,
        row_percentage: Optional[float] = None,
        before_transformation: Optional[bool] = None,
        training_dataset_version: Optional[int] = None,
    ) -> Optional[List[statistics.Statistics]]:
        """Get all statistics of an entity.

        :param metadata_instance: metadata object of the instance to get statistics of
        :type metadata_instance: TrainingDataset, FeatureGroup, FeatureView
        :param computation_time: Time at which statistics where computed
        :type computation_time: int
        :param start_commit_time: Window start commit time
        :type start_commit_time: int
        :param end_commit_time: Window end commit time
        :type end_commit_time: int
        :param feature_names: List of feature names of which statistics are retrieved
        :type feature_names: List[str]
        :param row_percentage: Percentage of feature values used during statistics computation
        :type row_percentage: float
        :param before_transformation: Whether the statistics were computed before transformations or not
        :type before_transformation: bool
        :param training_dataset_version: Version of the training dataset on which statistics were computed
        :type training_dataset_version: int
        """
        # get all statistics by entity + filters + sorts, without the feature descriptive statistics
        _client = client.get_instance()
        path_params = self.get_path(metadata_instance, training_dataset_version)

        # multiple statistics
        offset, limit = 0, None

        headers = {"content-type": "application/json"}
        query_params = self._build_get_query_params(
            computation_time=computation_time,
            start_commit_time=start_commit_time,
            end_commit_time=end_commit_time,
            filter_eq_times=False,
            feature_names=feature_names,
            row_percentage=row_percentage,
            before_transformation=before_transformation,
            training_dataset_version=training_dataset_version,
            # retrieve all entity statistics, excluding feature descriptive statistics
            offset=offset,
            limit=limit,
            with_content=False,
        )

        return statistics.Statistics.from_response_json(
            _client._send_request("GET", path_params, query_params, headers=headers)
        )

    def compute(
        self,
        metadata_instance: Union[
            training_dataset.TrainingDataset,
            feature_view.FeatureView,
            feature_group.FeatureGroup,
        ],
        training_dataset_version: Optional[int] = None,
    ) -> job.Job:
        """Compute statistics for an entity.

        :param metadata_instance: metadata object of the instance to compute statistics for
        :type metadata_instance: TrainingDataset, FeatureGroup, FeatureView
        :param training_dataset_version: version of the training dataset metadata object
        :type training_dataset_version: int
        """
        _client = client.get_instance()
        path_params = self.get_path(metadata_instance, training_dataset_version) + [
            "compute"
        ]
        return job.Job.from_response_json(_client._send_request("POST", path_params))

    def get_path(
        self,
        metadata_instance: Union[
            training_dataset.TrainingDataset,
            feature_group.FeatureGroup,
            feature_view.FeatureView,
        ],
        training_dataset_version: Optional[int] = None,
    ) -> List[Union[str, int]]:
        """Get statistics path.

        :param metadata_instance: metadata object of the instance to compute statistics for
        :type metadata_instance: TrainingDataset, FeatureGroup
        :param training_dataset_version: version of the training dataset metadata object
        :type training_dataset_version: int
        """
        _client = client.get_instance()
        if isinstance(metadata_instance, feature_view.FeatureView):
            path = [
                "project",
                _client._project_id,
                "featurestores",
                self._feature_store_id,
                "featureview",
                metadata_instance.name,
                "version",
                metadata_instance.version,
            ]
            if training_dataset_version is not None:
                path += [
                    "trainingdatasets",
                    "version",
                    training_dataset_version,
                ]
            return path + ["statistics"]
        else:
            return [
                "project",
                _client._project_id,
                "featurestores",
                self._feature_store_id,
                self._entity_type,
                metadata_instance.id,
                "statistics",
            ]

    def _extract_single_stats(
        self, stats: Union[statistics.Statistics, List[statistics.Statistics]]
    ) -> Optional[statistics.Statistics]:
        return stats[0] if isinstance(stats, list) else stats

    def _build_get_query_params(
        self,
        computation_time: Optional[int] = None,
        start_commit_time: Optional[int] = None,
        end_commit_time: Optional[int] = None,
        filter_eq_times: bool = False,
        feature_names: Optional[List[str]] = None,
        row_percentage: Optional[float] = None,
        before_transformation: Optional[bool] = None,
        training_dataset_version: Optional[int] = None,
        offset: int = 0,
        limit: Optional[int] = None,
        with_content: bool = False,
    ) -> Dict[str, Any]:
        """Build query parameters for statistics requests.

        :param computation_time: Time at which statistics where computed
        :type computation_time: int
        :param start_commit_time: Window start commit time
        :type start_commit_time: int
        :param end_commit_time: Window end commit time
        :type end_commit_time: int
        :param feature_names: List of feature names of which statistics are retrieved
        :type feature_names: List[str]
        :param row_percentage: Percentage of feature values used during statistics computation
        :type row_percentage: float
        :param before_transformation: Whether the statistics were computed for transformations or not
        :type before_transformation: bool
        :param training_dataset_version: Version of the training dataset on which statistics were computed
        :type training_dataset_version: int
        :param offset: Offset for pagination queries
        :type offset: int
        :param limit: Limit for pagination queries
        :type limit: int
        :param with_content: Whether include feature descriptive statistics in the response or not
        :type with_content: bool
        """
        query_params: dict[str, Union[int, str, List[str]]] = {"offset": offset}
        if limit is not None:
            query_params["limit"] = limit
        if with_content:
            query_params["fields"] = "content"

        sorts: List[str] = []
        filters: List[str] = []

        # filters and sorts

        # window times
        if end_commit_time is not None:
            col_name = "window_end_commit_time"
            sorts.append(col_name + ":desc")  # first sort
            filter_name = col_name + ("_eq" if filter_eq_times else "_ltoeq")
            filters.append(filter_name + ":" + str(end_commit_time))
        if start_commit_time is not None:
            col_name = "window_start_commit_time"
            sorts.append(col_name + ":asc")  # second sort
            filter_name = col_name + ("_eq" if filter_eq_times else "_gtoeq")
            filters.append(filter_name + ":" + str(start_commit_time))

        # computation time -- order is important, this should be after the other sorts
        if start_commit_time is None and end_commit_time is None:
            sorts.append("computation_time:desc")  # third sort
        if computation_time is not None:
            filters.append("computation_time_ltoeq:" + str(computation_time))

        # row percentage
        if row_percentage is not None:
            filters.append("row_percentage_eq:" + str(row_percentage))

        # for transformation
        if before_transformation is not None:
            filters.append("before_transformation_eq:" + str(before_transformation))

        # feature names
        if feature_names is not None:
            query_params["feature_names"] = feature_names

        # others
        if training_dataset_version is not None:
            query_params["training_dataset_version"] = training_dataset_version

        if sorts:
            query_params["sort_by"] = sorts
        if filters:
            query_params["filter_by"] = filters

        return query_params



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import functools
import os

from hopsworks_common.core.constants import (
    HAS_GREAT_EXPECTATIONS,
    HAS_POLARS,
    great_expectations_not_installed_message,
    polars_not_installed_message,
)


def not_connected(fn):
    @functools.wraps(fn)
    def if_not_connected(inst, *args, **kwargs):
        if inst._connected:
            raise HopsworksConnectionError
        return fn(inst, *args, **kwargs)

    return if_not_connected


def connected(fn):
    @functools.wraps(fn)
    def if_connected(inst, *args, **kwargs):
        if not inst._connected:
            raise NoHopsworksConnectionError
        return fn(inst, *args, **kwargs)

    return if_connected


class HopsworksConnectionError(Exception):
    """Thrown when attempted to change connection attributes while connected."""

    def __init__(self):
        super().__init__(
            "Connection is currently in use. Needs to be closed for modification."
        )


class NoHopsworksConnectionError(Exception):
    """Thrown when attempted to perform operation on connection while not connected."""

    def __init__(self):
        super().__init__(
            "Connection is not active. Needs to be connected for feature store operations."
        )


if os.environ.get("HOPSWORKS_RUN_WITH_TYPECHECK", False):
    from typeguard import typechecked
else:
    from typing import TypeVar

    _T = TypeVar("_T")

    def typechecked(
        target: _T,
    ) -> _T:
        return target if target else typechecked


def uses_great_expectations(f):
    @functools.wraps(f)
    def g(*args, **kwds):
        if not HAS_GREAT_EXPECTATIONS:
            raise ModuleNotFoundError(great_expectations_not_installed_message)
        return f(*args, **kwds)

    return g


def uses_polars(f):
    @functools.wraps(f)
    def g(*args, **kwds):
        if not HAS_POLARS:
            raise ModuleNotFoundError(polars_not_installed_message)
        return f(*args, **kwds)

    return g



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy
from hsml.utils.schema.tensor import Tensor


class TensorSchema:
    """Metadata object representing a tensor schema for a model."""

    def __init__(self, tensor_obj=None):
        if isinstance(tensor_obj, list):
            self.tensors = self._convert_list_to_schema(tensor_obj)
        elif isinstance(tensor_obj, numpy.ndarray):
            self.tensors = self._convert_tensor_to_schema(tensor_obj)
        else:
            raise TypeError(
                "{} is not supported in a tensor schema.".format(type(tensor_obj))
            )

    def _convert_tensor_to_schema(self, tensor_obj):
        return Tensor(tensor_obj.dtype, tensor_obj.shape)

    def _convert_list_to_schema(self, tensor_obj):
        if len(tensor_obj) == 1:
            return [self._build_tensor(tensor_obj[0])]
        else:
            tensors = []
            for tensor in tensor_obj:
                tensors.append(self._build_tensor(tensor))
        return tensors

    def _build_tensor(self, tensor_obj):
        name = None
        type = None
        shape = None
        description = None

        # Name is optional
        if "name" in tensor_obj:
            name = tensor_obj["name"]

        if "description" in tensor_obj:
            description = tensor_obj["description"]

        if "type" in tensor_obj:
            type = tensor_obj["type"]
        else:
            raise ValueError(
                "Mandatory 'type' key missing from entry {}".format(tensor_obj)
            )

        if "shape" in tensor_obj:
            shape = tensor_obj["shape"]
        else:
            raise ValueError(
                "Mandatory 'shape' key missing from entry {}".format(tensor_obj)
            )

        return Tensor(type, shape, name=name, description=description)



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Optional

from hsfs.core import expectation_api
from hsfs.ge_expectation import GeExpectation


class ExpectationEngine:
    def __init__(
        self, feature_store_id: int, feature_group_id: int, expectation_suite_id: int
    ):
        """Expectation engine.

        :param feature_store_id: id of the respective Feature Store
        :type feature_store_id: int
        :param feature_group_id: id of the respective Feature Group
        :type feature_group_id: int
        :param expectation_suite_id: id of the respective Expectation Suite
        :type expectation_suite_id: int
        """
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._expectation_suite_id = expectation_suite_id
        self._expectation_api = expectation_api.ExpectationApi(
            feature_store_id, feature_group_id, expectation_suite_id
        )

    # CRUD operations
    def create(self, expectation: GeExpectation) -> GeExpectation:
        return self._expectation_api.create(expectation)

    def update(self, expectation: GeExpectation) -> GeExpectation:
        return self._expectation_api.update(expectation)

    def get(self, expectation_id: int) -> Optional[GeExpectation]:
        return self._expectation_api.get(expectation_id)

    def delete(self, expectation_id: int) -> None:
        self._expectation_api.delete(expectation_id)

    # End of CRUD operations

    def get_expectations_by_suite_id(self):
        return self._expectation_api.get_expectations_by_suite_id()

    def check_for_id(self, expectation: GeExpectation) -> None:
        if expectation.id:
            return
        else:
            if "expectationId" in expectation.meta.keys():
                expectation.id = int(expectation.meta["expectationId"])
            else:
                raise ValueError(
                    "The provided expectation has no id. Either set the id field or populate the meta field with an expectationId keyword."
                )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Dict, List, Optional, Union

from hopsworks_common import usage
from hsml import client, util
from hsml import predictor as predictor_mod
from hsml.client.exceptions import ModelServingException
from hsml.client.istio.utils.infer_type import InferInput
from hsml.constants import DEPLOYABLE_COMPONENT, PREDICTOR_STATE
from hsml.core import model_api, serving_api
from hsml.engine import serving_engine
from hsml.inference_batcher import InferenceBatcher
from hsml.inference_logger import InferenceLogger
from hsml.predictor_state import PredictorState
from hsml.resources import Resources
from hsml.transformer import Transformer


class Deployment:
    """Metadata object representing a deployment in Model Serving."""

    def __init__(
        self,
        predictor,
        name: Optional[str] = None,
        description: Optional[str] = None,
        project_namespace: str = None,
        **kwargs,
    ):
        self._predictor = predictor
        self._description = description
        self._project_namespace = project_namespace

        if self._predictor is None:
            raise ModelServingException("A predictor is required")
        elif not isinstance(self._predictor, predictor_mod.Predictor):
            raise ValueError(
                "The predictor provided is not an instance of the Predictor class"
            )

        if name is not None:
            self._predictor.name = name

        if self._description is None:
            self._description = self._predictor.description
        else:
            self._description = self._predictor.description = description

        self._serving_api = serving_api.ServingApi()
        self._serving_engine = serving_engine.ServingEngine()
        self._model_api = model_api.ModelApi()
        self._grpc_channel = None
        self._model_registry_id = None

    @usage.method_logger
    def save(self, await_update: Optional[int] = 120):
        """Persist this deployment including the predictor and metadata to Model Serving.

        # Arguments
            await_update: If the deployment is running, awaiting time (seconds) for the running instances to be updated.
                          If the running instances are not updated within this timespan, the call to this method returns while
                          the update in the background.
        """

        self._serving_engine.save(self, await_update)

    @usage.method_logger
    def start(self, await_running: Optional[int] = 120):
        """Start the deployment

        # Arguments
            await_running: Awaiting time (seconds) for the deployment to start.
                           If the deployment has not started within this timespan, the call to this method returns while
                           it deploys in the background.
        """

        self._serving_engine.start(self, await_status=await_running)

    @usage.method_logger
    def stop(self, await_stopped: Optional[int] = 120):
        """Stop the deployment

        # Arguments
            await_stopped: Awaiting time (seconds) for the deployment to stop.
                           If the deployment has not stopped within this timespan, the call to this method returns while
                           it stopping in the background.
        """

        self._serving_engine.stop(self, await_status=await_stopped)

    @usage.method_logger
    def delete(self, force=False):
        """Delete the deployment

        # Arguments
            force: Force the deletion of the deployment.
                   If the deployment is running, it will be stopped and deleted automatically.
                   !!! warn A call to this method does not ask for a second confirmation.
        """

        self._serving_engine.delete(self, force)

    def get_state(self) -> PredictorState:
        """Get the current state of the deployment

        # Returns
            `PredictorState`. The state of the deployment.
        """

        return self._serving_engine.get_state(self)

    def is_created(self) -> bool:
        """Check whether the deployment is created.

        # Returns
            `bool`. Whether the deployment is created or not.
        """

        return (
            self._serving_engine.get_state(self).status
            != PREDICTOR_STATE.STATUS_CREATING
        )

    def is_running(self, or_idle=True, or_updating=True) -> bool:
        """Check whether the deployment is ready to handle inference requests

        # Arguments
            or_idle: Whether the idle state is considered as running (default is True)
            or_updating: Whether the updating state is considered as running (default is True)

        # Returns
            `bool`. Whether the deployment is ready or not.
        """

        status = self._serving_engine.get_state(self).status
        return (
            status == PREDICTOR_STATE.STATUS_RUNNING
            or (or_idle and status == PREDICTOR_STATE.STATUS_IDLE)
            or (or_updating and status == PREDICTOR_STATE.STATUS_UPDATING)
        )

    def is_stopped(self, or_created=True) -> bool:
        """Check whether the deployment is stopped

        # Arguments
            or_created: Whether the creating and created state is considered as stopped (default is True)

        # Returns
            `bool`. Whether the deployment is stopped or not.
        """

        status = self._serving_engine.get_state(self).status
        return status == PREDICTOR_STATE.STATUS_STOPPED or (
            or_created
            and (
                status == PREDICTOR_STATE.STATUS_CREATING
                or status == PREDICTOR_STATE.STATUS_CREATED
            )
        )

    def predict(
        self,
        data: Union[Dict, InferInput] = None,
        inputs: Union[List, Dict] = None,
    ):
        """Send inference requests to the deployment.
           One of data or inputs parameters must be set. If both are set, inputs will be ignored.

        !!! example
            ```python
            # login into Hopsworks using hopsworks.login()

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            # retrieve deployment by name
            my_deployment = ms.get_deployment("my_deployment")

            # (optional) retrieve model input example
            my_model = project.get_model_registry()  \
                              .get_model(my_deployment.model_name, my_deployment.model_version)

            # make predictions using model inputs (single or batch)
            predictions = my_deployment.predict(inputs=my_model.input_example)

            # or using more sophisticated inference request payloads
            data = { "instances": [ my_model.input_example ], "key2": "value2" }
            predictions = my_deployment.predict(data)
            ```

        # Arguments
            data: Payload dictionary for the inference request including the model input(s)
            inputs: Model inputs used in the inference requests

        # Returns
            `dict`. Inference response.
        """

        return self._serving_engine.predict(self, data, inputs)

    def get_model(self):
        """Retrieve the metadata object for the model being used by this deployment"""
        return self._model_api.get(
            self.model_name, self.model_version, self.model_registry_id
        )

    @usage.method_logger
    def download_artifact_files(self, local_path=None):
        """Download the artifact files served by the deployment

        # Arguments
            local_path: path where to download the artifact files in the local filesystem
        """

        return self._serving_engine.download_artifact_files(self, local_path=local_path)

    def get_logs(self, component="predictor", tail=10):
        """Prints the deployment logs of the predictor or transformer.

        # Arguments
            component: Deployment component to get the logs from (e.g., predictor or transformer)
            tail: Number of most recent lines to retrieve from the logs.
        """

        # validate component
        components = list(util.get_members(DEPLOYABLE_COMPONENT))
        if component not in components:
            raise ValueError(
                "Component '{}' is not valid. Possible values are '{}'".format(
                    component, ", ".join(components)
                )
            )

        logs = self._serving_engine.get_logs(self, component, tail)
        if logs is not None:
            for log in logs:
                print(log, end="\n\n")

    def get_url(self):
        """Get url to the deployment in Hopsworks"""

        path = (
            "/p/"
            + str(client.get_instance()._project_id)
            + "/deployments/"
            + str(self.id)
        )
        return util.get_hostname_replaced_url(path)

    def describe(self):
        """Print a description of the deployment"""

        util.pretty_print(self)

    @classmethod
    def from_response_json(cls, json_dict):
        predictors = predictor_mod.Predictor.from_response_json(json_dict)
        if isinstance(predictors, list):
            return [
                cls.from_predictor(predictor_instance)
                for predictor_instance in predictors
            ]
        else:
            return cls.from_predictor(predictors)

    @classmethod
    def from_predictor(cls, predictor_instance):
        return Deployment(
            predictor=predictor_instance,
            name=predictor_instance._name,
            description=predictor_instance._description,
        )

    def update_from_response_json(self, json_dict):
        self._predictor.update_from_response_json(json_dict)
        self.__init__(
            predictor=self._predictor,
            name=self._predictor._name,
            description=self._predictor._description,
        )
        return self

    def json(self):
        return self._predictor.json()

    def to_dict(self):
        return self._predictor.to_dict()

    # Deployment

    @property
    def id(self):
        """Id of the deployment."""
        return self._predictor.id

    @property
    def name(self):
        """Name of the deployment."""
        return self._predictor.name

    @name.setter
    def name(self, name: str):
        self._predictor.name = name

    @property
    def description(self):
        """Description of the deployment."""
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def predictor(self):
        """Predictor used in the deployment."""
        return self._predictor

    @predictor.setter
    def predictor(self, predictor):
        self._predictor = predictor

    @property
    def requested_instances(self):
        """Total number of requested instances in the deployment."""
        return self._predictor.requested_instances

    # Single predictor

    @property
    def model_name(self):
        """Name of the model deployed by the predictor"""
        return self._predictor.model_name

    @model_name.setter
    def model_name(self, model_name: str):
        self._predictor.model_name = model_name

    @property
    def model_path(self):
        """Model path deployed by the predictor."""
        return self._predictor.model_path

    @model_path.setter
    def model_path(self, model_path: str):
        self._predictor.model_path = model_path

    @property
    def model_version(self):
        """Model version deployed by the predictor."""
        return self._predictor.model_version

    @model_version.setter
    def model_version(self, model_version: int):
        self._predictor.model_version = model_version

    @property
    def artifact_version(self):
        """Artifact version deployed by the predictor."""
        return self._predictor.artifact_version

    @artifact_version.setter
    def artifact_version(self, artifact_version: Union[int, str]):
        self._predictor.artifact_version = artifact_version

    @property
    def artifact_files_path(self):
        """Path of the artifact files deployed by the predictor."""
        return self._predictor.artifact_files_path

    @property
    def artifact_path(self):
        """Path of the model artifact deployed by the predictor."""
        # TODO: deprecated
        return self._predictor.artifact_path

    @property
    def model_server(self):
        """Model server ran by the predictor."""
        return self._predictor.model_server

    @model_server.setter
    def model_server(self, model_server: str):
        self._predictor.model_server = model_server

    @property
    def serving_tool(self):
        """Serving tool used to run the model server."""
        return self._predictor.serving_tool

    @serving_tool.setter
    def serving_tool(self, serving_tool: str):
        self._predictor.serving_tool = serving_tool

    @property
    def script_file(self):
        """Script file used by the predictor."""
        return self._predictor.script_file

    @script_file.setter
    def script_file(self, script_file: str):
        self._predictor.script_file = script_file

    @property
    def resources(self):
        """Resource configuration for the predictor."""
        return self._predictor.resources

    @resources.setter
    def resources(self, resources: Resources):
        self._predictor.resources = resources

    @property
    def inference_logger(self):
        """Configuration of the inference logger attached to this predictor."""
        return self._predictor.inference_logger

    @inference_logger.setter
    def inference_logger(self, inference_logger: InferenceLogger):
        self._predictor.inference_logger = inference_logger

    @property
    def inference_batcher(self):
        """Configuration of the inference batcher attached to this predictor."""
        return self._predictor.inference_batcher

    @inference_batcher.setter
    def inference_batcher(self, inference_batcher: InferenceBatcher):
        self._predictor.inference_batcher = inference_batcher

    @property
    def transformer(self):
        """Transformer configured in the predictor."""
        return self._predictor.transformer

    @transformer.setter
    def transformer(self, transformer: Transformer):
        self._predictor.transformer = transformer

    @property
    def model_registry_id(self):
        """Model Registry Id of the deployment."""
        return self._model_registry_id

    @model_registry_id.setter
    def model_registry_id(self, model_registry_id: int):
        self._model_registry_id = model_registry_id

    @property
    def created_at(self):
        """Created at date of the predictor."""
        return self._predictor.created_at

    @property
    def creator(self):
        """Creator of the predictor."""
        return self._predictor.creator

    @property
    def api_protocol(self):
        """API protocol enabled in the deployment (e.g., HTTP or GRPC)."""
        return self._predictor.api_protocol

    @api_protocol.setter
    def api_protocol(self, api_protocol: str):
        self._predictor.api_protocol = api_protocol

    @property
    def environment(self):
        """Name of inference environment"""
        return self._predictor.environment

    @environment.setter
    def environment(self, environment: str):
        self._predictor.environment = environment

    @property
    def project_namespace(self):
        """Name of inference environment"""
        return self._predictor.project_namespace

    @project_namespace.setter
    def project_namespace(self, project_namespace: str):
        self._predictor.project_namespace = project_namespace

    def __repr__(self):
        desc = (
            f", description: {self._description!r}"
            if self._description is not None
            else ""
        )
        return f"Deployment(name: {self._predictor._name!r}" + desc + ")"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from abc import ABC, abstractmethod
from typing import Optional, Union

import humps
from hsml import util
from hsml.constants import Default
from hsml.inference_batcher import InferenceBatcher
from hsml.resources import Resources


class DeployableComponent(ABC):
    """Configuration of a deployable component (predictor or transformer)."""

    def __init__(
        self,
        script_file: Optional[str] = None,
        resources: Optional[Resources] = None,
        inference_batcher: Optional[Union[InferenceBatcher, dict, Default]] = None,
        **kwargs,
    ):
        self._script_file = script_file
        self._resources = resources
        self._inference_batcher = (
            util.get_obj_from_json(inference_batcher, InferenceBatcher)
            or InferenceBatcher()
        )

    @classmethod
    @abstractmethod
    def from_json(cls, json_decamelized):
        "To be implemented by the component type"
        pass

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls.from_json(json_decamelized)

    def json(self):
        return json.dumps(self, cls=util.MLEncoder)

    @abstractmethod
    def update_from_response_json(self, json_dict):
        "To be implemented by the component type"
        pass

    @abstractmethod
    def to_dict(self):
        "To be implemented by the component type"
        pass

    @property
    def script_file(self):
        """Script file ran by the deployment component (i.e., predictor or transformer)."""
        return self._script_file

    @script_file.setter
    def script_file(self, script_file: str):
        self._script_file = script_file

    @property
    def resources(self):
        """Resource configuration for the deployment component (i.e., predictor or transformer)."""
        return self._resources

    @resources.setter
    def resources(self, resources: Resources):
        self._resources = resources

    @property
    def inference_batcher(self):
        """Configuration of the inference batcher attached to the deployment component (i.e., predictor or transformer)."""
        return self._inference_batcher

    @inference_batcher.setter
    def inference_batcher(self, inference_batcher: InferenceBatcher):
        self._inference_batcher = inference_batcher



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import pytest
from hsml import (
    inference_batcher,
    inference_logger,
    predictor,
    resources,
    transformer,
    util,
)
from hsml.constants import MODEL, PREDICTOR, RESOURCES


SERVING_RESOURCE_LIMITS = {"cores": 2, "memory": 1024, "gpus": 2}
SERVING_NUM_INSTANCES_NO_LIMIT = [-1]
SERVING_NUM_INSTANCES_SCALE_TO_ZERO = [0]
SERVING_NUM_INSTANCES_ONE = [0]


class TestPredictor:
    # from response json

    def test_from_response_json_empty(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        json = backend_fixtures["predictor"]["get_deployments_empty"]["response"]

        # Act
        pred = predictor.Predictor.from_response_json(json)

        # Assert
        assert isinstance(pred, list)
        assert len(pred) == 0

    def test_from_response_json_singleton(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        json = backend_fixtures["predictor"]["get_deployments_singleton"]["response"]

        # Act
        pred = predictor.Predictor.from_response_json(json)

        # Assert
        assert isinstance(pred, list)
        assert len(pred) == 1

        p = pred[0]
        p_json = json["items"][0]

        assert isinstance(p, predictor.Predictor)
        assert p.id == p_json["id"]
        assert p.name == p_json["name"]
        assert p.description == p_json["description"]
        assert p.created_at == p_json["created"]
        assert p.creator == p_json["creator"]
        assert p.model_path == p_json["model_path"]
        assert p.model_name == p_json["model_name"]
        assert p.model_version == p_json["model_version"]
        assert p.model_framework == p_json["model_framework"]
        assert p.model_server == p_json["model_server"]
        assert p.serving_tool == p_json["serving_tool"]
        assert p.api_protocol == p_json["api_protocol"]
        assert p.artifact_version == p_json["artifact_version"]
        assert p.environment == p_json["environment_dto"]["name"]
        assert p.script_file == p_json["predictor"]
        assert isinstance(p.resources, resources.PredictorResources)
        assert isinstance(p.transformer, transformer.Transformer)
        assert p.transformer.script_file == p_json["transformer"]
        assert isinstance(p.transformer.resources, resources.TransformerResources)
        assert isinstance(p.inference_logger, inference_logger.InferenceLogger)
        assert p.inference_logger.mode == p_json["inference_logging"]
        assert isinstance(p.inference_batcher, inference_batcher.InferenceBatcher)
        assert p.inference_batcher.enabled == bool(
            p_json["batching_configuration"]["batching_enabled"]
        )

    def test_from_response_json_list(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        json = backend_fixtures["predictor"]["get_deployments_list"]["response"]

        # Act
        pred = predictor.Predictor.from_response_json(json)

        # Assert
        assert isinstance(pred, list)
        assert len(pred) == 2

        for i in range(len(pred)):
            p = pred[i]
            p_json = json["items"][i]

            assert isinstance(p, predictor.Predictor)
            assert p.id == p_json["id"]
            assert p.name == p_json["name"]
            assert p.description == p_json["description"]
            assert p.created_at == p_json["created"]
            assert p.creator == p_json["creator"]
            assert p.model_path == p_json["model_path"]
            assert p.model_name == p_json["model_name"]
            assert p.model_version == p_json["model_version"]
            assert p.model_framework == p_json["model_framework"]
            assert p.model_server == p_json["model_server"]
            assert p.serving_tool == p_json["serving_tool"]
            assert p.api_protocol == p_json["api_protocol"]
            assert p.environment == p_json["environment_dto"]["name"]
            assert p.artifact_version == p_json["artifact_version"]
            assert p.script_file == p_json["predictor"]
            assert isinstance(p.resources, resources.PredictorResources)
            assert isinstance(p.transformer, transformer.Transformer)
            assert p.transformer.script_file == p_json["transformer"]
            assert isinstance(p.transformer.resources, resources.TransformerResources)
            assert isinstance(p.inference_logger, inference_logger.InferenceLogger)
            assert p.inference_logger.mode == p_json["inference_logging"]
            assert isinstance(p.inference_batcher, inference_batcher.InferenceBatcher)
            assert p.inference_batcher.enabled == bool(
                p_json["batching_configuration"]["batching_enabled"]
            )

    def test_from_response_json_single(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        p_json = backend_fixtures["predictor"]["get_deployments_singleton"]["response"][
            "items"
        ][0]

        # Act
        p = predictor.Predictor.from_response_json(p_json)

        # Assert
        assert isinstance(p, predictor.Predictor)
        assert p.id == p_json["id"]
        assert p.name == p_json["name"]
        assert p.description == p_json["description"]
        assert p.created_at == p_json["created"]
        assert p.creator == p_json["creator"]
        assert p.model_path == p_json["model_path"]
        assert p.model_version == p_json["model_version"]
        assert p.model_name == p_json["model_name"]
        assert p.model_framework == p_json["model_framework"]
        assert p.model_server == p_json["model_server"]
        assert p.serving_tool == p_json["serving_tool"]
        assert p.api_protocol == p_json["api_protocol"]
        assert p.environment == p_json["environment_dto"]["name"]
        assert p.artifact_version == p_json["artifact_version"]
        assert p.script_file == p_json["predictor"]
        assert isinstance(p.resources, resources.PredictorResources)
        assert isinstance(p.transformer, transformer.Transformer)
        assert p.transformer.script_file == p_json["transformer"]
        assert isinstance(p.transformer.resources, resources.TransformerResources)
        assert isinstance(p.inference_logger, inference_logger.InferenceLogger)
        assert p.inference_logger.mode == p_json["inference_logging"]
        assert isinstance(p.inference_batcher, inference_batcher.InferenceBatcher)
        assert p.inference_batcher.enabled == bool(
            p_json["batching_configuration"]["batching_enabled"]
        )

    # constructor

    def test_constructor(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        p_json = backend_fixtures["predictor"]["get_deployments_singleton"]["response"][
            "items"
        ][0]
        mock_validate_serving_tool = mocker.patch(
            "hsml.predictor.Predictor._validate_serving_tool",
            return_value=p_json["serving_tool"],
        )
        mock_resources = util.get_obj_from_json(
            copy.deepcopy(p_json["predictor_resources"]), resources.PredictorResources
        )
        mock_validate_resources = mocker.patch(
            "hsml.predictor.Predictor._validate_resources",
            return_value=mock_resources,
        )
        mock_validate_script_file = mocker.patch(
            "hsml.predictor.Predictor._validate_script_file",
            return_value=p_json["predictor"],
        )

        # Act
        p = predictor.Predictor(
            id=p_json["id"],
            name=p_json["name"],
            description=p_json["description"],
            created_at=p_json["created"],
            creator=p_json["creator"],
            model_path=p_json["model_path"],
            model_version=p_json["model_version"],
            model_name=p_json["model_name"],
            model_framework=p_json["model_framework"],
            model_server=p_json["model_server"],
            serving_tool=p_json["serving_tool"],
            api_protocol=p_json["api_protocol"],
            environment=p_json["environment_dto"]["name"],
            artifact_version=p_json["artifact_version"],
            script_file=p_json["predictor"],
            resources=p_json["predictor_resources"],
            transformer={
                "script_file": p_json["transformer"],
                "resources": copy.deepcopy(p_json["transformer_resources"]),
            },
            inference_logger={
                "mode": p_json["inference_logging"],
                "kafka_topic": copy.deepcopy(p_json["kafka_topic_dto"]),
            },
            inference_batcher=copy.deepcopy(p_json["batching_configuration"]),
        )

        # Assert
        assert p.id == p_json["id"]
        assert p.name == p_json["name"]
        assert p.description == p_json["description"]
        assert p.created_at == p_json["created"]
        assert p.creator == p_json["creator"]
        assert p.model_path == p_json["model_path"]
        assert p.model_name == p_json["model_name"]
        assert p.model_version == p_json["model_version"]
        assert p.model_framework == p_json["model_framework"]
        assert p.model_server == p_json["model_server"]
        assert p.serving_tool == p_json["serving_tool"]
        assert p.api_protocol == p_json["api_protocol"]
        assert p.environment == p_json["environment_dto"]["name"]
        assert p.artifact_version == p_json["artifact_version"]
        assert p.script_file == p_json["predictor"]
        assert isinstance(p.resources, resources.PredictorResources)
        assert isinstance(p.transformer, transformer.Transformer)
        assert p.transformer.script_file == p_json["transformer"]
        assert isinstance(p.transformer.resources, resources.TransformerResources)
        assert isinstance(p.inference_logger, inference_logger.InferenceLogger)
        assert p.inference_logger.mode == p_json["inference_logging"]
        assert isinstance(p.inference_batcher, inference_batcher.InferenceBatcher)
        assert p.inference_batcher.enabled == bool(
            p_json["batching_configuration"]["batching_enabled"]
        )
        mock_validate_serving_tool.assert_called_once_with(p_json["serving_tool"])
        assert mock_validate_resources.call_count == 1
        mock_validate_script_file.assert_called_once_with(
            p_json["model_framework"], p_json["predictor"]
        )

    # validate serving tool

    def test_validate_serving_tool_none(self):
        # Act
        st = predictor.Predictor._validate_serving_tool(None)

        # Assert
        assert st is None

    def test_validate_serving_tool_valid(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, is_saas_connection=False
        )

        # Act
        st = predictor.Predictor._validate_serving_tool(PREDICTOR.SERVING_TOOL_DEFAULT)

        # Assert
        assert st == PREDICTOR.SERVING_TOOL_DEFAULT

    def test_validate_serving_tool_invalid(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, is_saas_connection=False
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            _ = predictor.Predictor._validate_serving_tool("INVALID_NAME")

        # Assert
        assert "is not valid" in str(e_info.value)

    def test_validate_serving_tool_valid_saas_connection(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, is_saas_connection=True
        )

        # Act
        st = predictor.Predictor._validate_serving_tool(PREDICTOR.SERVING_TOOL_KSERVE)

        # Assert
        assert st == PREDICTOR.SERVING_TOOL_KSERVE

    def test_validate_serving_tool_invalid_saas_connection(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, is_saas_connection=True
        )

        # Act
        with pytest.raises(ValueError) as e_info:
            _ = predictor.Predictor._validate_serving_tool(
                PREDICTOR.SERVING_TOOL_DEFAULT
            )

        # Assert
        assert "KServe deployments are the only supported" in str(e_info.value)

    # validate script file

    def test_validate_script_file_tf_none(self):
        # Act
        predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_TENSORFLOW, None)

    def test_validate_script_file_sk_none(self):
        # Act
        predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_SKLEARN, None)

    def test_validate_script_file_th_none(self):
        # Act
        predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_TORCH, None)

    def test_validate_script_file_py_none(self):
        # Act
        with pytest.raises(ValueError) as e_info:
            _ = predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_PYTHON, None)

        # Assert
        assert "Predictor scripts are required" in str(e_info.value)

    def test_validate_script_file_llm_none(self):
        # Act
        with pytest.raises(ValueError) as e_info:
            _ = predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_LLM, None)

        # Assert
        assert "Predictor scripts are required" in str(e_info.value)

    def test_validate_script_file_tf_script_file(self):
        # Act
        predictor.Predictor._validate_script_file(
            MODEL.FRAMEWORK_TENSORFLOW, "script_file"
        )

    def test_validate_script_file_sk_script_file(self):
        # Act
        predictor.Predictor._validate_script_file(
            MODEL.FRAMEWORK_SKLEARN, "script_file"
        )

    def test_validate_script_file_th_script_file(self):
        # Act
        predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_TORCH, "script_file")

    def test_validate_script_file_py_script_file(self):
        # Act
        predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_PYTHON, "script_file")

    def test_validate_script_file_llm_script_file(self):
        # Act
        predictor.Predictor._validate_script_file(MODEL.FRAMEWORK_LLM, "script_file")

    # infer model server

    def test_infer_model_server_tf(self):
        # Act
        ms = predictor.Predictor._infer_model_server(MODEL.FRAMEWORK_TENSORFLOW)

        # Assert
        assert ms == PREDICTOR.MODEL_SERVER_TF_SERVING

    def test_infer_model_server_sk(self):
        # Act
        ms = predictor.Predictor._infer_model_server(MODEL.FRAMEWORK_SKLEARN)

        # Assert
        assert ms == PREDICTOR.MODEL_SERVER_PYTHON

    def test_infer_model_server_th(self):
        # Act
        ms = predictor.Predictor._infer_model_server(MODEL.FRAMEWORK_TORCH)

        # Assert
        assert ms == PREDICTOR.MODEL_SERVER_PYTHON

    def test_infer_model_server_py(self):
        # Act
        ms = predictor.Predictor._infer_model_server(MODEL.FRAMEWORK_PYTHON)

        # Assert
        assert ms == PREDICTOR.MODEL_SERVER_PYTHON

    def test_infer_model_server_llm(self):
        # Act
        ms = predictor.Predictor._infer_model_server(MODEL.FRAMEWORK_LLM)

        # Assert
        assert ms == PREDICTOR.MODEL_SERVER_VLLM

    # default serving tool

    def test_get_default_serving_tool_kserve_installed(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, is_kserve_installed=True
        )

        # Act
        st = predictor.Predictor._get_default_serving_tool()

        # Assert
        assert st == PREDICTOR.SERVING_TOOL_KSERVE

    def test_get_default_serving_tool_kserve_not_installed(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, is_kserve_installed=False
        )

        # Act
        st = predictor.Predictor._get_default_serving_tool()

        # Assert
        assert st == PREDICTOR.SERVING_TOOL_DEFAULT

    # validate resources

    def test_validate_resources_none_non_kserve(self):
        # Act
        res = predictor.Predictor._validate_resources(
            None, PREDICTOR.SERVING_TOOL_DEFAULT
        )

        # Assert
        assert res is None

    def test_validate_resources_none_kserve(self):
        # Act
        res = predictor.Predictor._validate_resources(
            None, PREDICTOR.SERVING_TOOL_KSERVE
        )

        # Assert
        assert res is None

    def test_validate_resources_num_instances_zero_non_kserve(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        pr = resources.PredictorResources(num_instances=0)

        # Act
        res = predictor.Predictor._validate_resources(
            pr, PREDICTOR.SERVING_TOOL_DEFAULT
        )

        # Assert
        assert res == pr

    def test_validate_resources_num_instances_zero_kserve(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        pr = resources.PredictorResources(num_instances=0)

        # Act
        res = predictor.Predictor._validate_resources(pr, PREDICTOR.SERVING_TOOL_KSERVE)

        # Assert
        assert res == pr

    def test_validate_resources_num_instances_one_without_scale_to_zero_non_kserve(
        self, mocker
    ):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        pr = resources.PredictorResources(num_instances=1)

        # Act
        res = predictor.Predictor._validate_resources(
            pr, PREDICTOR.SERVING_TOOL_DEFAULT
        )

        # Assert
        assert res == pr

    def test_validate_resources_num_instances_one_without_scale_to_zero_kserve(
        self, mocker
    ):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )
        pr = resources.PredictorResources(num_instances=1)

        # Act
        res = predictor.Predictor._validate_resources(pr, PREDICTOR.SERVING_TOOL_KSERVE)

        # Assert
        assert res == pr

    def test_validate_resources_num_instances_one_with_scale_to_zero_non_kserve(
        self, mocker
    ):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )
        pr = resources.PredictorResources(num_instances=1)

        # Act
        res = predictor.Predictor._validate_resources(
            pr, PREDICTOR.SERVING_TOOL_DEFAULT
        )

        # Assert
        assert res == pr

    def test_validate_resources_num_instances_one_with_scale_to_zero_kserve(
        self, mocker
    ):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )
        pr = resources.PredictorResources(num_instances=1)

        # Act
        with pytest.raises(ValueError) as e_info:
            _ = predictor.Predictor._validate_resources(
                pr, PREDICTOR.SERVING_TOOL_KSERVE
            )

        # Assert
        assert "Scale-to-zero is required" in str(e_info.value)

    # default resources

    def test_get_default_resources_non_kserve_without_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )

        # Act
        res = predictor.Predictor._get_default_resources(PREDICTOR.SERVING_TOOL_DEFAULT)

        # Assert
        assert isinstance(res, resources.PredictorResources)
        assert res.num_instances == RESOURCES.MIN_NUM_INSTANCES

    def test_get_default_resources_non_kserve_with_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )

        # Act
        res = predictor.Predictor._get_default_resources(PREDICTOR.SERVING_TOOL_DEFAULT)

        # Assert
        assert isinstance(res, resources.PredictorResources)
        assert res.num_instances == RESOURCES.MIN_NUM_INSTANCES

    def test_get_default_resources_kserve_without_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=False
        )

        # Act
        res = predictor.Predictor._get_default_resources(PREDICTOR.SERVING_TOOL_KSERVE)

        # Assert
        assert isinstance(res, resources.PredictorResources)
        assert res.num_instances == RESOURCES.MIN_NUM_INSTANCES

    def test_get_default_resources_kserve_with_scale_to_zero(self, mocker):
        # Arrange
        self._mock_serving_variables(
            mocker, SERVING_NUM_INSTANCES_NO_LIMIT, force_scale_to_zero=True
        )

        # Act
        res = predictor.Predictor._get_default_resources(PREDICTOR.SERVING_TOOL_KSERVE)

        # Assert
        assert isinstance(res, resources.PredictorResources)
        assert res.num_instances == 0

    # for model

    def test_for_model(self, mocker):
        # Arrange
        def spec(model, model_name, model_version, model_path):
            pass

        mock_get_predictor_for_model = mocker.patch(
            "hsml.util.get_predictor_for_model", return_value=True, spec=spec
        )

        class MockModel:
            name = "model_name"
            version = "model_version"
            model_path = "model_path"

        mock_model = MockModel()

        # Act
        predictor.Predictor.for_model(mock_model)

        # Assert
        mock_get_predictor_for_model.assert_called_once_with(
            model=mock_model,
            model_name=mock_model.name,
            model_version=mock_model.version,
            model_path=mock_model.model_path,
        )

    # extract fields from json

    def extract_fields_from_json(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        p_json = backend_fixtures["predictor"]["get_deployments_singleton"]["response"][
            "items"
        ][0]

        # Act
        kwargs = predictor.Predictor.extract_fields_from_json(p_json)

        # Assert
        assert kwargs["id"] == p_json["id"]
        assert kwargs["name"] == p_json["name"]
        assert kwargs["description"] == p_json["description"]
        assert kwargs["created_at"] == p_json["created"]
        assert kwargs["creator"] == p_json["creator"]
        assert kwargs["model_name"] == p_json["model_name"]
        assert kwargs["model_path"] == p_json["model_path"]
        assert kwargs["model_version"] == p_json["model_version"]
        assert kwargs["model_framework"] == p_json["model_framework"]
        assert kwargs["artifact_version"] == p_json["artifact_version"]
        assert kwargs["model_server"] == p_json["model_server"]
        assert kwargs["serving_tool"] == p_json["serving_tool"]
        assert kwargs["script_file"] == p_json["predictor"]
        assert isinstance(kwargs["resources"], resources.PredictorResources)
        assert isinstance(kwargs["inference_logger"], inference_logger.InferenceLogger)
        assert kwargs["inference_logger"].mode == p_json["inference_logging"]
        assert isinstance(
            kwargs["inference_batcher"], inference_batcher.InferenceBatcher
        )
        assert kwargs["inference_batcher"].enabled == bool(
            p_json["batching_configuration"]["batching_enabled"]
        )
        assert kwargs["api_protocol"] == p_json["api_protocol"]
        assert kwargs["environment"] == p_json["environment_dto"]["name"]
        assert isinstance(kwargs["transformer"], transformer.Transformer)
        assert kwargs["transformer"].script_file == p_json["transformer"]
        assert isinstance(
            kwargs["transformer"].resources, resources.TransformerResources
        )

    # deploy

    def test_deploy(self, mocker, backend_fixtures):
        # Arrange
        self._mock_serving_variables(mocker, SERVING_NUM_INSTANCES_NO_LIMIT)
        p_json = backend_fixtures["predictor"]["get_deployments_singleton"]["response"][
            "items"
        ][0]
        mock_deployment_init = mocker.patch(
            "hsml.deployment.Deployment.__init__", return_value=None
        )
        mock_deployment_save = mocker.patch("hsml.deployment.Deployment.save")

        # Act

        p = predictor.Predictor.from_response_json(p_json)
        p.deploy()

        # Assert
        mock_deployment_init.assert_called_once_with(
            predictor=p,
            name=p.name,
            description=p.description,
        )
        mock_deployment_save.assert_called_once()

    # auxiliary methods

    def _mock_serving_variables(
        self,
        mocker,
        num_instances,
        force_scale_to_zero=False,
        is_saas_connection=False,
        is_kserve_installed=True,
    ):
        mocker.patch(
            "hsml.client.get_serving_resource_limits",
            return_value=SERVING_RESOURCE_LIMITS,
        )
        mocker.patch(
            "hsml.client.get_serving_num_instances_limits", return_value=num_instances
        )
        mocker.patch(
            "hsml.client.is_scale_to_zero_required", return_value=force_scale_to_zero
        )
        mocker.patch("hsml.client.is_saas_connection", return_value=is_saas_connection)
        mocker.patch(
            "hsml.client.is_kserve_installed", return_value=is_kserve_installed
        )
